# flake8: noqa
recipe_without_positive_iso3_code = (
    '''{
      	"mapnumber": "MA001",
       	"category": "Reference",
        "product": "{e.country_name}: Overview Map",
       	"summary": "Overview of {e.country_name} with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap_stle_stl_pt_s0_allmaps",
                        "reg_exp": "^{e.affected_country_iso3}_stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?).shp$",
                        "schema_definition": "stle_ste_pt.yml",
                        "definition_query": "'NAME_0' = '{e.country_name}'",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": [
                            {
                                "class_name": "National Capital",
                                "expression": "[name] (Capital of {e.country_name})",
                                "sql_query": "('fclass' = 'national_capital')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Admin 1 Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'town')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Nieghbouring Country Capitals",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'national_capital') AND ('ADM0_Pcode' <> '{e.affected_country_iso3}')",
                                "show_class_labels": true
                            }                            
                        ]
                    }
                ]
            }
        ]
    }'''
)

recipe_without_negative_iso3_code = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "{e.country_name}: Overview Map",
       	"summary": "Overview of {e.country_name} with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap-admn-ad1-py-s0-reference",
                        "reg_exp": "^(?!({e.affected_country_iso3}))_admn_ad0_py_(.*?)_(.*?)_([phm][phm])(.+)shp$",
                        "schema_definition": "admin1_reference.yml",
                        "definition_query": "ADM0_NAME <> '{e.country_name}'",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": []
                    }
                ]
            }
        ]
    }'''
)

recipe_with_layer_name_only = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "{e.country_name}: Overview Map",
       	"summary": "Overview of {e.country_name} with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap-admn-ad1-py-s0-reference"
                    }
                ]
            }
        ]
    }'''
)

recipe_with_non_standard_principal_map_frame_name = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "{e.country_name}: Overview Map",
       	"summary": "Overview of {e.country_name} with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "My favourite map frame",
        "map_frames": [
            {
                "name": "My favourite map frame",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap-admn-ad1-py-s0-reference"
                    }
                ]
            }
        ]
    }'''
)


recipe_with_invalid_principal_map_frame_name = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "{e.country_name}: Overview Map",
       	"summary": "Overview of {e.country_name} with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "This map frame does not exist",
        "map_frames": [
            {
                "name": "The map frame",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap-admn-ad1-py-s0-reference"
                    }
                ]
            }
        ]
    }'''
)

recipe_schema_v2_0_with_layer_name_only = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "{e.country_name}: Overview Map",
       	"summary": "Overview of {e.country_name} with topography displayed",
    	"export": true,
        "template": "reference",
        "map_frames": [
            {
                "name": "Main map",
                "layers": [
                    {
                        "name": "mainmap-admn-ad1-py-s0-reference"
                    }
                ]
            }
        ]
    }'''
)

recipe_with_layer_details_embedded = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "{e.country_name}: Overview Map",
       	"summary": "Overview of {e.country_name} with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap-admn-ad1-py-s0-reference",
                        "reg_exp": "^[a-z]{3}_stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?).shp$",
                        "schema_definition": "admin1_reference.yml",
                        "definition_query": "fclass IN ('national_capital', 'city', 'capital', 'town')",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": [
                            {
                                "class_name": "National Capital",
                                "expression": "[name]",
                                "sql_query": "(\\"fclass\\" = 'national_capital')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Admin 1 Capital",
                                "expression": "[name]",
                                "sql_query": "(\\"fclass\\" = 'town')",
                                "show_class_labels": true
                            }
                        ]
                    }
                ]
            }
        ]
    }'''
)


recipe_with_positive_iso3_code = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "Mozambique: Overview Map",
       	"summary": "Overview of Mozambique with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap_stle_stl_pt_s0_allmaps",
                        "reg_exp": "^moz_stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?).shp$",
                        "schema_definition": "stle_ste_pt.yml",
                        "definition_query": "'NAME_0' = 'Mozambique'",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": [
                            {
                                "class_name": "National Capital",
                                "expression": "[name] (Capital of Mozambique)",
                                "sql_query": "('fclass' = 'national_capital')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Admin 1 Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'town')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Nieghbouring Country Capitals",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'national_capital') AND ('ADM0_Pcode' <> 'moz')",
                                "show_class_labels": true
                            }                                  
                        ]
                    }
                ]
            }
        ]
    }'''
)


recipe_with_negative_iso3_code = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "Mozambique: Overview Map",
       	"summary": "Overview of Mozambique with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap-admn-ad1-py-s0-reference",
                        "reg_exp": "^(?!(moz))_admn_ad0_py_(.*?)_(.*?)_([phm][phm])(.+)shp$",
                        "schema_definition": "admin1_reference.yml",
                        "definition_query": "ADM0_NAME <> 'Mozambique'",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": []
                    }
                ]
            }
        ]
    }'''
)


recipe_test_for_search_for_shapefiles = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "Mozambique: Overview Map",
       	"summary": "Overview of Mozambique with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap_stle_stl_pt_s0_allmaps",
                        "reg_exp": "^moz_stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?).shp$",
                        "schema_definition": "stle_ste_pt.yml",
                        "definition_query": "fclass IN ('national_capital', 'city', 'capital', 'town')",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": [
                            {
                                "class_name": "National Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'national_capital')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Admin 1 Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'town')",
                                "show_class_labels": true
                            }
                        ]
                    }
                ]
            }
        ]
    }'''
)


recipe_result_one_dataset_per_layer_windows = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "Mozambique: Overview Map",
       	"summary": "Overview of Mozambique with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap_stle_stl_pt_s0_allmaps",
                        "reg_exp": "^moz_stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?).shp$",
                        "schema_definition": "stle_ste_pt.yml",
                        "data_source_path": "D:/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_stle_ste_pt_s0_osm_pp.shp",
                        "data_name": "moz_stle_ste_pt_s0_osm_pp",
                        "definition_query": "fclass IN ('national_capital', 'city', 'capital', 'town')",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": [
                            {
                                "class_name": "National Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'national_capital')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Admin 1 Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'town')",
                                "show_class_labels": true
                            }
                        ]
                    }
                ]
            }
        ]
    }'''
)

recipe_result_one_dataset_per_layer_linux = (
    '''{
      	"mapnumber": "MA001",
    	"category": "Reference",
        "product": "Mozambique: Overview Map",
       	"summary": "Overview of Mozambique with topography displayed",
    	"export": true,
        "template": "reference",
        "principal_map_frame": "Main map",
        "map_frames": [
            {
                "name": "Main map",
				"crs": "EPSG:3857",
                "layers": [
                    {
                        "name": "mainmap_stle_stl_pt_s0_allmaps",
                        "reg_exp": "^moz_stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?).shp$",
                        "schema_definition": "stle_ste_pt.yml",
                        "data_source_path": "/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_stle_ste_pt_s0_osm_pp.shp",
                        "data_name": "moz_stle_ste_pt_s0_osm_pp",
                        "definition_query": "fclass IN ('national_capital', 'city', 'capital', 'town')",
                        "display": true,
                        "add_to_legend": true,
                        "label_classes": [
                            {
                                "class_name": "National Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'national_capital')",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Admin 1 Capital",
                                "expression": "[name]",
                                "sql_query": "('fclass' = 'town')",
                                "show_class_labels": true
                            }
                        ]
                    }
                ]
            }
        ]
    }'''
)


recipe_with_invalid_layer_file_path = (
    '''{
        "category": "Reference",
        "product": "Mozambique: Overview Map",
        "principal_map_frame": "Main map",
        "mapnumber": "MA001",
        "summary": "Overview of Mozambique with topography displayed",
        "export": true,
        "template": "reference",
        "map_frames": [
            {
                "layers": [
                    {
                        "data_source_path": "D://MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_stle_ste_pt_s0_osm_pp.shp",
                        "name": "mainmap_stle_stl_pt_s0_allmaps",
                        "schema_definition": "stle_ste_pt.yml",
                        "data_schema": {
                            "required": [
                                "name_en"
                            ],
                            "properties": {
                                "geometry_type": {
                                    "items": {
                                        "enum": [
                                            "MultiPolygon",
                                            "Polygon"
                                        ]
                                    },
                                    "additionalItems": false
                                }
                            }
                        },
                        "data_name": "moz_stle_ste_pt_s0_osm_pp",
                        "definition_query": "fclass IN ('national_capital', 'city', 'capital', 'town')",
                        "label_classes": [
                            {
                                "class_name": "National Capital",
                                "sql_query": "('fclass' = 'national_capital')",
                                "expression": "[name]",
                                "show_class_labels": true
                            },
                            {
                                "class_name": "Admin 1 Capital",
                                "sql_query": "('fclass' = 'town')",
                                "expression": "[name]",
                                "show_class_labels": true
                            }
                        ],
                        "layer_file_path": "D:/code/github/mapactionpy_controller/mapactionpy_controller/example/mainmap_stle_stl_pt_s0_allmaps",
                        "data_source_checksum": "d41d8cd98f00b204e9800998ecf8427e",
                        "add_to_legend": true,
                        "display": true,
                        "reg_exp": "^moz_stle_ste_pt_(.*?)_(.*?)_([phm][phm])(.*?).shp$"
                    }
                ],
                "crs": "EPSG:3857",
                "name": "Main map"
            }
        ]
    }'''
)

#     '''{
#         "title": "{e.affectedcountry}: Overview Map",
#     "layers": [
#    {
#        "map_frame": "Main Map",
#        "layer_group": "Admin - Polygons",
#        "layer_display_name": "Admin - AffectedCountry - py",
#        "search_definition": "^moz_admn_ad0_py_(.*?)_(.*?)_([phm][phm])(.+)shp$",
#        "data_source_path": "D:/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_py_s0_unknown_pp.shp",
#        "data_name": "moz_admn_ad0_py_s0_unknown_pp",
#        "rendering": "Admin - AffectedCountry - py",
#        "definition_query": "None",
#        "visible": "Yes"
#    }
#     ]
# }'''
# )

fixture_datasource_query = r"""
{
"settlement_points": "^{e.affected_country_iso3}_stle_stl_pt_(.*?)_(.*?)_([phm][phm])(_(.+))",
"surrounding_counties": "^(?!({e.affected_country_iso3}))_admn_ad0_py_(.*?)_(.*?)_([phm][phm])(_(.+))"
"airports_points": "D:\MapAction\2019-06-12-GBR\GIS\2_Active_Data\232_tran\scr_tran_air_pt_s1_ourairports_pp.shp"
}
"""

fixture_datasource_intermediatory_query = r"""
{
"settlement_points": "^gbr_stle_stl_pt_(.*?)_(.*?)_([phm][phm])(_(.+))",
"surrounding_counties": "^(?!(gbr))_admn_ad0_py_(.*?)_(.*?)_([phm][phm])(_(.+))"
"airports_points": "D:\MapAction\2019-06-12-GBR\GIS\2_Active_Data\232_tran\scr_tran_air_pt_s1_ourairports_pp.shp"
}
"""

fixture_datasource_result_one_dataset_per_layer = r"""
{
"settlement_points": "D:\MapAction\2019-06-12-GBR\GIS\2_Active_Data\gbr_stle_stle_pt_s0_naturalearth_pp.shp",
"airports_points": "D:\MapAction\2019-06-12-GBR\GIS\2_Active_Data\232_tran\scr_tran_air_pt_s1_ourairports_pp.shp"
}
"""

fixture_datasource_result_missing_layer = r"""
{
"settlement_points": "",
"airports_points": "D:\MapAction\2019-06-12-GBR\GIS\2_Active_Data\232_tran\scr_tran_air_pt_s1_ourairports_pp.shp"
}
"""

walk_single_admn_file_search_search = \
    [
        (
            "D:/MapAction/2019MOZ01/GIS/2_Active_Data",
            ['202_admn'],
            ['desktop.ini']
        ),
        (
            "D:/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn",
            [],
            ['moz_stle_ste_pt_s0_osm_pp.shp',
             'moz_stle_ste_pt_s0_osm_pp.dbf',
             'moz_stle_ste_pt_s0_osm_pp.sbx',
             'moz_stle_ste_pt_s0_osm_pp.sbn',
             'moz_stle_ste_pt_s0_osm_pp.prj',
             'moz_admn_ad0_ln_s0_unknown_pp.CPG',
             'moz_admn_ad0_ln_s0_unknown_pp.dbf',
             'moz_admn_ad0_py_s0_unknown_pp.sbx',
             'moz_admn_ad0_py_s0_unknown_pp.sbn',
             'moz_admn_ad0_py_s0_unknown_pp.prj',
             'moz_admn_ad0_py_s0_unknown_pp.dbf',
             'moz_admn_ad0_ln_s0_unknown_pp.shp',
             'moz_admn_ad0_ln_s0_unknown_pp.shx',
             'moz_admn_ad0_py_s0_unknown_pp.CPG',
             'moz_admn_ad0_ln_s0_unknown_pp.sbx',
             'moz_admn_ad0_ln_s0_unknown_pp.prj',
             'moz_admn_ad0_ln_s0_unknown_pp.sbn',
             'moz_admn_ad1_py_s1_mapaction_pp.shx',
             'moz_admn_ad1_ln_s1_mapaction_pp.shp.xml',
             'moz_admn_ad1_ln_s1_mapaction_pp.prj',
             'moz_admn_ad1_ln_s1_mapaction_pp.sbx',
             'moz_admn_ad1_ln_s1_mapaction_pp.sbn',
             'moz_admn_ad1_ln_s1_mapaction_pp.shp.LAPTOP-F7PICP7J.8304.13164.sr.lock',
             'moz_admn_ad1_py_s1_mapaction_pp.shp.xml',
             'moz_admn_ad1_py_s1_mapaction_pp.sbx',
             'moz_admn_ad1_py_s1_mapaction_pp.sbn',
             'moz_admn_ad1_py_s1_mapaction_pp.shp.LAPTOP-F7PICP7J.8304.13164.sr.lock',
             'moz_admn_ad1_py_s1_mapaction_pp.prj',
             'moz_admn_ad1_ln_s1_mapaction_pp.shx',
             'moz_admn_ad1_py_s1_mapaction_pp.dbf',
             'moz_admn_ad1_ln_s1_mapaction_pp.dbf',
             'moz_admn_ad0_py_s0_unknown_pp.shx',
             'moz_admn_ad1_ln_s1_mapaction_pp.CPG',
             'moz_admn_ad0_py_s0_unknown_pp.shp',
             'moz_admn_ad0_py_s0_unknown_pp.shp.xml',
             'moz_admn_ad1_py_s1_mapaction_pp.CPG',
             'moz_admn_ad1_py_s1_mapaction_pp.shp',
             'moz_admn_ad1_ln_s1_mapaction_pp.shp',
             'desktop.ini']
        )
    ]

glob_single_stle_file_search_windows = [
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_stle_ste_pt_s0_osm_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad0_ln_s0_unknown_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad0_py_s0_unknown_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad1_py_s1_mapaction_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad1_ln_s1_mapaction_pp.shp',
]

glob_single_stle_file_search_linux = [
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_stle_ste_pt_s0_osm_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_ln_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_py_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_py_s1_mapaction_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_ln_s1_mapaction_pp.shp',
]

glob_multiple_stle_file_search_windows = [
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_stle_ste_pt_s0_osm_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_stle_ste_pt_s0_gmapmaker_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad0_ln_s0_unknown_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad0_py_s0_unknown_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad1_py_s1_mapaction_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad1_ln_s1_mapaction_pp.shp',
]

glob_multiple_stle_file_search_linux = [
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_stle_ste_pt_s0_osm_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_stle_ste_pt_s0_gmapmaker_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_ln_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_py_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_py_s1_mapaction_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_ln_s1_mapaction_pp.shp',
]

glob_multiple_stle_file_search_case_difference_linux = [
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_stle_ste_pt_s0_osm_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_STLE_ste_pt_s0_osm_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_ln_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_py_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_py_s1_mapaction_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_ln_s1_mapaction_pp.shp',
]


glob_no_stle_file_search_windows = [
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad0_ln_s0_unknown_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad0_py_s0_unknown_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad1_py_s1_mapaction_pp.shp',
    'D:\\MapAction\\2019MOZ01\\GIS\\2_Active_Data\\202_admn\\moz_admn_ad1_ln_s1_mapaction_pp.shp',
]

glob_no_stle_file_search_linux = [
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_ln_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad0_py_s0_unknown_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_py_s1_mapaction_pp.shp',
    '/user/home/MapAction/2019MOZ01/GIS/2_Active_Data/202_admn/moz_admn_ad1_ln_s1_mapaction_pp.shp',
]



"""
# main_stack

The main stack within `mapaction_controller` exists in two modules; this one and steps.py. Each individual
step is represented by a Step object, which is a wrapper for:

* A function to run (which must accept kwargs**).
* A definition of a "running", "success" and "failure" message.
* A severity of failure.

The main function in main_stack.py is `process_stack()`, which works through a list of "Step" objects. For
each Step:

    * Its function is called.
    * The function may return:
        - One or more new Step objects (which are inserted into the slack). Closures are very useful here
        - An updated state object (which will passed to the next step)
    * Each Step defines each of a "running", "success" and "failure" message, which is displayed on the
      console and/or log as appropriate.


Execution continous until there are no more Steps.

Note: main_stack.py also includes humanfriendly.terminal which ought really to be part of the CLI.
"""
import logging
import traceback
import six
import sys
from collections import deque

import humanfriendly.terminal as hft
import humanfriendly.terminal.spinners as spinners

from mapactionpy_controller.steps import Step
from mapactionpy_controller.task_renderer import TaskReferralBase

logger = logging.getLogger(__name__)

bright_white = hft.ansi_style(color='white', bright=True)
bright_green = hft.ansi_style(color='green', bright=True)
bright_red = hft.ansi_style(color='red', bright=True)
bright_yellow = hft.ansi_style(color='yellow', bright=True)
normal_white = hft.ansi_style(color='white', bright=False)

terminal_checkboxs = {
    logging.INFO:  '{}[{}pass{}]{}'.format(normal_white, bright_green, normal_white, bright_white),
    logging.ERROR: '{}[{}fail{}]{}'.format(normal_white, bright_red, normal_white, bright_white),
    logging.WARNING: '{}[{}warn{}]{}'.format(normal_white, bright_yellow, normal_white, bright_white)
}


def get_jira_client():
    try:
        from mapactionpy_controller.jira_tasks import JiraClient
        return JiraClient()
    except (ImportError, ValueError):
        return None


jira_client = get_jira_client()


def parse_feedback(status, msg, step, **kwargs):
    """
    This is called once per step execution.
    It provides a hook into print messages to the terminal, log files and the JIRA Client.
    """
    # nice_output = '{}\n{}'.format(self.complete_msg, result)
    # pass_back['result'] = result
    # pass_back['exp'] = exp

    the_msg = msg
    task_referal = None
    if status > logging.INFO:
        exp = kwargs['exp']
        if exp.args:
            if isinstance(exp.args[0], TaskReferralBase):
                task_referal = exp.args[0]
                # error_str = task_referal.get_task_unique_summary()
            else:
                error_str = '\n'.join(str(s) for s in exp.args if isinstance(s, six.string_types))

                stack_trace = kwargs['stack_trace']
                # error_str = '\n'.join(str(s) for s in exp.args if isinstance(s, six.string_types))
                the_msg = '{}\nerror message={}\n\n{}'.format(
                    msg, error_str, stack_trace)

    if jira_client:
        jira_client.task_handler(status, msg, task_referal)

    if hft.connected_to_terminal():
        hft.output('{} {} {}'.format(
            hft.ANSI_ERASE_LINE,
            terminal_checkboxs[status],
            the_msg)
        )
    else:
        logger.log(status, the_msg)


def _add_steps_from_state_to_stack(new_state, stack, old_state):
    """
    This helper function checks to see Step has returned one or more addtional Steps. If so these
    are added to the top of the stack.

    :param new_state: The return value of the most recently called Step object.
    :param stack: The stack.
    :param old_state: The state value which was passed to the most recently called Step object.
    :returns: If `new_state` contains Step objects then `old_state` is returned. Else `new_state`
       is returned
    """
    if isinstance(new_state, Step):
        stack.append(new_state)
        return old_state

    if isinstance(new_state, list) and all([isinstance(stp, Step) for stp in new_state]):
        new_state.reverse()
        stack.extend(new_state)
        return old_state

    return new_state


def process_stack(step_list, initial_state):
    """
    This is the principal function which executes the stack of `Step` objects. For each step in the stack
    its `func` is called.
    * If `func` returns one or more Step objects (either as a single object or as a list) then these are
      added to the top of the stack. The `state` remains unaltered and is passed to the next Step.
    * If `func` returns any other value (including None and other falsey values) then this is passed as
      the state object to the `func` of the next Step item in the stack.

    :param step_list: A list of initial steps which are used to populate the stack.
    :param initial_state: This value will be passed a the 'state' keyword arg to the first step's `func`.
    :returns: The return value of the final step's `func`.
    """
    hft.enable_ansi_support()
    n_state = initial_state
    step_list.reverse()
    stack = deque(step_list)

    try:
        while stack:
            # Definitions:
            # `n_state` = the state for the current iteration
            # `nplus_state` = the state for the next iteraction (eg N+1)
            step = stack.pop()
            kwargs = {'state': n_state}

            if hft.connected_to_terminal():
                with spinners.AutomaticSpinner(step.running_msg, show_time=True):
                    nplus_state = step.run(parse_feedback, **kwargs)
            else:
                logger.info('Starting: {}'.format(step.running_msg))
                nplus_state = step.run(parse_feedback, **kwargs)

            # Used to increment the state *only* if no new Steps where returned
            n_state = _add_steps_from_state_to_stack(nplus_state, stack, n_state)

        return n_state
    except Exception as exp:
        pass_back = {
            'exp': exp,
            'stack_trace': traceback.format_exc()
        }
        # print error and then exit with a non-zero exit code
        parse_feedback(logging.ERROR, 'Unable to continue following the previous error', None, **pass_back)
        sys.exit(1)



from mapactionpy_controller.map_recipe import MapRecipe
import json


class MapCookbook:
    """
    MapCookbook - Contains recipes for Map Products
    """

    def __init__(self, cmf, layer_props, verify_on_creation=True, hum_event=None):
        """
        Sets path for Map Cookbook json file.
        Creates empty list of of products.

        Positional Arguments:
        @param cmf {CrashMoveFolder}: a CrashMoveFolder object
        @param layer_props {LayerProperties}: a LayerProperties object

        Optional Named Arguments:
        @param verify_on_creation {bool}: If True (default) then contents of the `cmf.map_definitions`
            json file are compared to the `layer_props`. If the `cmf.map_definitions` contains references to
            layers which are not described by the `layer_props` object then a ValueError will be raised.

        @param hum_event {Event}: An Event object. If not `None`, this is passed to the underlying recipe
                                  objects so that it can be access easily whilst they are processed.

        """
        self._check_cmf_param(cmf, layer_props, verify_on_creation)
        self.products = {}
        self.layer_props = layer_props
        self.hum_event = hum_event
        self._parse_json_file()

        if verify_on_creation:
            msg = self.layer_props.get_difference_with_other_layer_set(
                self.get_all_included_lyrs_as_set(),
                self._get_mismatch_wtih_lyr_props_message
            )
            if msg:
                raise ValueError(msg)

    def _check_cmf_param(self, cmf, layer_props, verify_on_creation):
        if cmf.verify_paths():
            self.cookbook_json_file = cmf.map_definitions
        else:
            raise ValueError('The `cmf` parameter for MapCookbook.__init__() can only accept'
                             ' values where the paths verify. eg `cmf.verify_paths() == True`.'
                             ' The value passed in this case failed this test')

        if verify_on_creation and not (layer_props.cmf.layer_properties == cmf.layer_properties):
            raise ValueError('Attempting to create a MapCookBook using a a CMF object and LayerProperties'
                             ' object which point to different layer_properties.json files. This is probably'
                             ' not what you want and may produce strange results. If you are sure this is want'
                             ' you require, then use `verify_on_creation=False` in the MapCookBook constructor.\n'
                             ' Values passed to the MapCookBook constructor\n'
                             '   cmf.layer_properties={}\n'
                             '   layer_props={}\n'.format(
                                 cmf.layer_properties,
                                 layer_props.cmf.layer_properties
                             ))

    def _parse_json_file(self):
        """
        Reads product "recipes" from Map Cookbook json file
        """
        with open(self.cookbook_json_file) as json_file:
            jsonContents = json.load(json_file)
            for recipe in jsonContents['recipes']:
                rec = MapRecipe(recipe, self.layer_props, self.hum_event)
                self.products[recipe['product']] = rec

    def get_all_included_lyrs_as_set(self):
        cb_unique_lyrs = set()

        for recipe in self.products.values():
            cb_unique_lyrs.update(recipe.get_lyrs_as_set())

        return cb_unique_lyrs

    def _get_mismatch_wtih_lyr_props_message(self, lp_only, cb_only):
        msg = ('There is a mismatch between the layer_properties.json file:\n\t"{}"\n'
               'and the MapCookbook.json file:\n\t"{}"\n'
               'One or more layer names occur in only one of these files.\n'.format(
                   self.layer_props.cmf.layer_properties,
                   self.cookbook_json_file
               ))

        pair = ((cb_only, 'These layers are only mentioned in the MapCookbook json file and not in Layer'
                          ' Properties json file:'),
                (lp_only, 'These layers are only mentioned in the Layer Properties json file and not in the'
                          ' MapCookbook json file:'))

        return self.layer_props._msg_builder(pair, msg)



from unittest import TestCase
import mapactionpy_controller.event as event
import os
from jsonschema import ValidationError
import json
import six


class TestEvent(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(
            os.path.dirname(os.path.realpath(__file__)))
        self.event_descriptor_path = os.path.join(
            self.parent_dir, 'example', 'event_description.json')
        self.failing_event_descriptor_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_event_description_missing_operation_id.json')

    def test_constructor(self):
        # test creation when using a valid json description:
        test_cmf = event.Event(self.event_descriptor_path)
        self.assertIsInstance(test_cmf, event.Event)

        # test exception is raised when passed a non-existant file
        self.assertRaises(IOError, event.Event, '/path/to/nonexistant/file.json')
        # test exception is raised for an invalid json file:
        self.assertRaises(ValidationError, event.Event, self.failing_event_descriptor_path)

    def test_fictional_country_name(self):
        with open(self.event_descriptor_path, 'r') as edf:
            base_event_def = json.load(edf)

        working_name_supplied = [
            ('ABC', u'Alphabet Land'),
            ('ABC', u'UNKNOWN'),
            ('GEO', u'Georgia'),
            ('MOZ', u'Mozambique'),
            ('COG', u'Congo'),
            ('COD', u'Congo'),
            ('GBR', u'United Kingdom'),
            ('GBR', u'England'),
            ('GBR', u'Ireland'),
            ('IRL', u'Ireland')
        ]

        for test_iso3, test_name in working_name_supplied:
            # Do supply the country name
            event_def = base_event_def.copy()
            event_def['affected_country_iso3'] = test_iso3
            event_def['country_name'] = test_name
            actual_name = event._parse_country_name(event_def)
            self.assertEqual(test_name, actual_name)

        working_name_not_supplied = [
            ('COG', u'Congo'),
            ('COD', u'Congo, The Democratic Republic of the'),
            ('GBR', u'United Kingdom'),
            ('IRL', u'Ireland')
        ]

        for test_iso3, test_name in working_name_not_supplied:
            # Don't supply the country name
            event_def = base_event_def.copy()
            event_def['affected_country_iso3'] = test_iso3
            actual_name = event._parse_country_name(event_def)
            self.assertEqual(test_name, actual_name)

        error_cases = [
            ('MOZ', u'Alphabet Land', 'supply a real ISO3 code with a fictional country name'),
            ('ABC', u'Mozambique', 'supply a real country name with a fictional ISO3 code'),
            ('ABC', None, 'A `country_name` value must be specified for fictional countries')
        ]

        for test_iso3, test_name, fail_msg in error_cases:
            event_def = base_event_def.copy()
            event_def['affected_country_iso3'] = test_iso3
            event_def['country_name'] = test_name
            with self.assertRaises(ValueError) as ve:
                event._parse_country_name(event_def)

            if six.PY2:
                self.assertRegexpMatches(str(ve.exception), fail_msg)
            else:
                self.assertRegex(str(ve.exception), fail_msg)



import json
import os
from unittest import TestCase
import fixtures
from jsonschema import ValidationError
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.event import Event
from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.map_recipe import MapRecipe
import jsonpickle
import six
# works differently for python 2.7 and python 3.x
try:
    from unittest import mock
except ImportError:
    import mock


class TestMAController(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.cmf_descriptor_path = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        self.cmf = CrashMoveFolder(self.cmf_descriptor_path, verify_on_creation=False)
        self.event_descriptor_path = os.path.join(self.parent_dir, 'example', 'event_description.json')
        self.event = Event(self.event_descriptor_path)
        # self.cmf = CrashMoveFolder(self.cmf_descriptor_path)
        self.lyr_props = LayerProperties(self.cmf, '', verify_on_creation=False)

        self.recipe_descriptor_path = os.path.join(self.parent_dir, 'example', 'example_single_map_recipe.json')
        with open(self.recipe_descriptor_path) as json_file:
            self.recipe_def = json.load(json_file)

        self.recipe = MapRecipe(self.recipe_def, self.lyr_props)

    def test_equality_of_map_recipes(self):
        recipe_def = json.loads(fixtures.recipe_with_positive_iso3_code)
        test_recipe1 = MapRecipe(recipe_def, self.lyr_props)
        test_recipe2 = MapRecipe(recipe_def, self.lyr_props)

        self.assertEqual(test_recipe1, test_recipe2)
        test_recipe1.summary = "something different"
        self.assertNotEqual(test_recipe1, test_recipe2)

        test_recipe3 = MapRecipe(recipe_def, self.lyr_props)
        self.assertEqual(test_recipe2, test_recipe3)
        # Remove an arbitary layer from each Map Frame
        for m_frames in test_recipe3.map_frames:
            m_frames.layers.pop()
        self.assertNotEqual(test_recipe2, test_recipe3)

    def test_serialise_and_deserialise_map_recipe(self):
        recipes_fixtures = [
            fixtures.recipe_with_positive_iso3_code,
            fixtures.recipe_result_one_dataset_per_layer_windows,
            fixtures.recipe_without_positive_iso3_code
        ]

        with mock.patch('mapactionpy_controller.map_recipe.path.exists') as mock_path:
            mock_path.return_value = True

            for fixture_str in recipes_fixtures:
                test_recipe = MapRecipe(fixture_str, self.lyr_props)

                self.assertEqual(
                    test_recipe,
                    jsonpickle.decode(jsonpickle.encode(test_recipe))
                )
                self.assertEqual(
                    test_recipe,
                    MapRecipe(jsonpickle.encode(test_recipe, unpicklable=False), self.lyr_props)
                )
                self.assertNotEqual(
                    test_recipe,
                    MapRecipe(fixtures.recipe_with_negative_iso3_code, self.lyr_props)
                )

    def test_cmf_schema_validation(self):

        test_files = ('fixture_cmf_description_extra_field.json',
                      'fixture_cmf_description_missing_field.json')

        for test_f in test_files:
            cmf_partial_fail = os.path.join(self.parent_dir, 'tests', 'testfiles', test_f)
            self.assertRaises(ValidationError, CrashMoveFolder, cmf_partial_fail)

    def test_cmf_path_validation(self):

        cmf_partial_fail = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cmf_description_one_file_and_one_dir_not_valid.json')

        # test validation on creation (for failing case) using default parameters
        self.assertRaises(ValueError, CrashMoveFolder, cmf_partial_fail)
        # force validation on creation, for a json file we know would otherwise fail:
        self.assertRaises(ValueError, CrashMoveFolder, cmf_partial_fail, verify_on_creation=True)
        # test validation is correctly disabled on creation, for a json file we know would otherwise fail:
        test_cmf = CrashMoveFolder(cmf_partial_fail, verify_on_creation=False)
        self.assertIsInstance(test_cmf, CrashMoveFolder)

        # check message included in the ValueError:
        with self.assertRaises(ValueError) as cm:
            test_cmf = CrashMoveFolder(cmf_partial_fail, verify_on_creation=True)

        if six.PY2:
            self.assertRegexpMatches(str(cm.exception), "map_templates")
            self.assertNotRegexpMatches(str(cm.exception), "original_data")
        else:
            self.assertRegex(str(cm.exception), "map_templates")
            self.assertNotRegex(str(cm.exception), "original_data")

        # create a valid CMF object and then test paths, after creation
        test_cmf_path = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        test_cmf = CrashMoveFolder(test_cmf_path)
        self.assertTrue(test_cmf.verify_paths())
        test_cmf_path = os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json')
        test_cmf = CrashMoveFolder(test_cmf_path)
        self.assertTrue(test_cmf.verify_paths())
        test_cmf.active_data = os.path.join(self.parent_dir, 'DOES-NOT-EXIST')
        self.assertFalse(test_cmf.verify_paths())



import csv
import os
from collections import namedtuple

import six

import mapactionpy_controller as mac

# abstract class
# Done using the "old-school" method described here, without using the abs module
# https://stackoverflow.com/a/25300153


class NamingClause(object):
    def __init__(self, nc_json_path, ** kwargs):
        if self.__class__ is NamingClause:
            raise NotImplementedError(
                'NamingClause is an abstract class and cannot be instantiated directly')

    def validate(self, clause_value, **kwargs):
        # if self.__class__ is NamingClause:
        raise NotImplementedError(
            'NamingClause is an abstract class and the `validate` method cannot be called directly')


class NamingFreeTextClause(NamingClause):
    def __init__(self, nc_json_path, ** kwargs):
        self.clause_name = kwargs['name']
        self.alias = kwargs['alias']

    def validate(self, clause_value):
        details = {self.alias: clause_value}

        class DataClauseValues(namedtuple('DataClauseValues', details.keys())):
            __slots__ = ()

            @property
            def is_valid(self):
                return True

            @property
            def get_message(self):
                return '"{}" is valid for freetext (as is almost anything)'.format(clause_value)

        return DataClauseValues(**details)


class NamingLookupClause(NamingClause):
    def __init__(self, nc_json_path, **kwargs):
        self.clause_name = kwargs['name']
        self.known_values = {}
        lookup_field = kwargs['lookup_field']
        nc_lookup_dir = os.path.dirname(nc_json_path)
        self.csv_filename = kwargs['filename']
        self.csv_filepath = os.path.join(nc_lookup_dir, self.csv_filename)

        if six.PY2:
            with open(self.csv_filepath, 'rb') as csv_file:
                self._init_known_values(self.csv_filepath, csv_file, lookup_field)
        else:
            with open(self.csv_filepath, 'r', newline='', encoding='iso-8859-1') as csv_file:
                self._init_known_values(self.csv_filepath, csv_file, lookup_field)

    def _init_known_values(self, csv_path, csv_file, lookup_field):
        csv_reader = csv.DictReader(
            csv_file, delimiter=',', quotechar='"')

        if lookup_field in csv_reader.fieldnames:
            self.lookup_field = lookup_field
        else:
            raise mac.name_convention.NamingException(
                'invalid validation lookup_field primary key "{}" in file "{}"'.format(lookup_field, csv_path))

        for row in csv_reader:
            pk = row[lookup_field].lower()
            if pk not in self.known_values:
                non_lookup_keys = [x for x in row.keys() if x != lookup_field]
                self.known_values[pk] = {n: row[n] for n in non_lookup_keys}
            else:
                raise mac.name_convention.NamingException(
                    'Duplicate primary key "{}" in file "{}"'.format(pk, csv_path))

    def validate(self, clause_value):
        clause_value = clause_value.lower()
        if clause_value in self.known_values.keys():
            details = self.known_values[clause_value]
            details[self.lookup_field] = clause_value
            valid_value = True
            message = '"{}" is a recognised value for the clause "{}" found in "{}"'.format(
                clause_value,
                self.clause_name,
                self.csv_filename
            )
        else:
            # print("{}".format(clause_value))
            details = {self.lookup_field: clause_value}
            valid_value = False
            message = '"{}" is not a recognised value for the clause "{}" found in "{}"'.format(
                clause_value,
                self.clause_name,
                self.csv_filename
            )

        class DataClauseValues(namedtuple('DataClauseValues', details.keys())):
            __slots__ = ()

            @property
            def is_valid(self):
                return valid_value

            @property
            def get_message(self):
                return message

        # DataClauseValues = namedtuple('DataClauseValues', details.keys())
        return DataClauseValues(**details)



import json
import logging
from os import path

import jsonpickle
import jsonschema

import mapactionpy_controller.state_serialization as state_serialization
from mapactionpy_controller import _get_validator_for_config_schema
from mapactionpy_controller.recipe_frame import RecipeFrame
from mapactionpy_controller.recipe_atlas import RecipeAtlas


logger = logging.getLogger(__name__)
validate_against_recipe_schema_v0_3 = _get_validator_for_config_schema('map-recipe-v0.3.schema')
validate_against_recipe_schema_v0_2 = _get_validator_for_config_schema('map-recipe-v0.2.schema')


class MapRecipe:
    """
    MapRecipe
    """
    OPTIONAL_FIELDS = ('atlas', 'hum_event', 'map_project_path', 'runners', 'template_path', 'version_num')

    def __init__(self, recipe_definition, lyr_props, hum_event=None):
        if isinstance(recipe_definition, dict):
            recipe_def = recipe_definition
        else:
            recipe_def = json.loads(recipe_definition)

        compatiblity_mode = self._check_schemas_with_backward_compat(recipe_def)

        # Required fields
        self.mapnumber = recipe_def["mapnumber"]
        self.category = recipe_def["category"]
        self.export = recipe_def["export"]
        self.product = recipe_def["product"]
        # This is a list, but see note in `_parse_map_frames` method
        self.map_frames = self._parse_map_frames(recipe_def["map_frames"], lyr_props, compatiblity_mode)
        self.summary = recipe_def["summary"]
        self.template = recipe_def["template"]
        self.principal_map_frame = self._parse_principal_map_frame(recipe_def, compatiblity_mode)

        # Optional fields
        self.hum_event = hum_event
        self.map_project_path = recipe_def.get('map_project_path', None)
        if self.map_project_path:
            self.map_project_path = path.abspath(self.map_project_path)
        self.template_path = recipe_def.get('template_path', None)
        self.version_num = recipe_def.get('version_num', None)
        self.runners = recipe_def.get('runners', None)
        atlas_def = recipe_def.get('atlas', None)
        if atlas_def:
            self.atlas = RecipeAtlas(atlas_def, self, lyr_props)
        else:
            self.atlas = None

        # Self consistency checks
        self._check_for_dup_text_elements()
        # print(self)

    def _check_schemas_with_backward_compat(self, recipe_def):
        try:
            validate_against_recipe_schema_v0_3(recipe_def)
            return 0.3
        except jsonschema.ValidationError as ve_v0_3:
            try:
                validate_against_recipe_schema_v0_2(recipe_def)
                # Do something useful here
                # Hack some values? Or raise a JIRA ticket?
                logger.warn('Attempting to load backwards compatable v0.2 MapRecipe')
                # raise ValueError('old maprecipe format')
                return 0.2
            except jsonschema.ValidationError:
                raise ve_v0_3

    def get_lyrs_as_set(self):
        # this is required for the case that the lyr is a str of the layername
        # This can only happen if the layername was not found in the Layerproperties file
        def get_lyr_name(lyr):
            try:
                return lyr.name
            except AttributeError:
                return lyr

        unique_lyrs = set()
        for mf in self.map_frames:

            lyrs = [get_lyr_name(lyr) for lyr in mf.layers]
            unique_lyrs.update(lyrs)

        return unique_lyrs

    def _parse_map_frames(self, map_frames_def, lyr_props, compatiblity_mode=0.3):
        # We create a seperate list nad set here so that we can enforce unique map_frames names. However only
        # the list is returned. Client code is generally more readable and elegant if `self.map_frames` is a
        # list. This enforces that map_frames names must be unique in the json representation, however
        # theoretically allows client code to create multiple map frames with identical names. The behaviour
        # in this circumstance is not known or tested and is entirely the client's responsiblity.
        recipe_map_frames_list = []
        map_frames_set = set()
        for frame_def in map_frames_def:
            mf = RecipeFrame(frame_def, lyr_props, compatiblity_mode)
            if mf.name in map_frames_set:
                raise ValueError(
                    'Duplicate mapframe name {} in recipe {}. Each mapframe name within a'
                    ' recipe must unique'.format(mf.name, self.product))

            map_frames_set.add(mf.name)
            recipe_map_frames_list.append(mf)

        return recipe_map_frames_list

    def _parse_principal_map_frame(self, recipe_def, compatiblity_mode):
        """
        This assumes that `_parse_map_frames` as already been called.
        """
        # The hard coded value "Main map" is for legacy reasons (recipe schema v0.2)
        p_map_frame = recipe_def.get('principal_map_frame', "Main map")

        if p_map_frame not in [mf.name for mf in self.map_frames]:
            err_msg = (
                'Unable to find a MapFrame "{}" in the recipe. The `principal_map_frame` value must have the'
                ' name of one of the MapFram objects in the recipe.')

            if compatiblity_mode == 0.2:
                err_msg = ('Unable to find a MapFrame "Main map" in the recipe. Please update the recipe to'
                           ' v0.3 format or later.')

            raise ValueError(err_msg)

        return p_map_frame

    def _check_for_dup_text_elements(self):
        # check that any named `scale_text_element`s and `spatial_ref_text_element`s
        # are each only referred to by a single map_frame
        scale_text_elements_set = set()
        spatial_ref_text_elements_set = set()

        for mf in self.map_frames:
            self._find_dups(mf.scale_text_element, scale_text_elements_set,
                            'The Map Recipe definition is invalid. More than one "map_frame" is linked to the'
                            ' Scale text element "{}"'
                            )
            self._find_dups(mf.spatial_ref_text_element, spatial_ref_text_elements_set,
                            'The Map Recipe definition is invalid. More than one "map_frame" is linked to the'
                            ' Spatial reference text element "{}"'
                            )

    def _find_dups(self, elem, aggregate_set, msg):
        if elem:
            if elem in aggregate_set:
                raise ValueError(msg.format(elem))
            else:
                aggregate_set.add(elem)

    def contains_frame(self, requested_frame_name):
        """
        Check whether or not an atlas with the given name exists in the MapRecipe.
        Returns a boolean
        """
        return requested_frame_name in [mf.name for mf in self.map_frames]

    def get_frame(self, requested_frame_name):
        """
        Gets an atlas by name.
        @returns the RecipeAtlas object
        @raises ValueError if the requested_atlas_name does not exist
        """
        # We trust that the map frame names are unique
        try:
            return [mf for mf in self.map_frames if mf.name == requested_frame_name][0]
        except IndexError:
            raise ValueError(
                'The requested map frame {} does not exist in the recipe {}'.format(
                    requested_frame_name, self.product))

    def all_layers(self):
        """
        A convenience method which all layers within the MapRecipe, irrespective of the MapFrame. Saves the
        caller needing to iterate through the MapFrames.

        @returns A list of every layer reference in the Map from all MapFrames.
        """
        # Doubtless there is a neater, comprehension based way of doing this.
        result = []
        for mf in self.map_frames:
            result.extend(mf.layers)

        # print('MapRecipe.all_layers() = {}'.format(result))
        return result
        # return list(itertools.chain([mf.layers for mf in self.map_frames]))

    def __str__(self):
        return json.dumps(json.loads(jsonpickle.encode(self, unpicklable=False)), indent=4)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        return not self.__eq__(other)

    def __getstate__(self):
        return state_serialization.get_state_optional_fields(self, MapRecipe.OPTIONAL_FIELDS)

    def __setstate__(self, state):
        state_serialization.set_state_optional_fields(self, state, MapRecipe.OPTIONAL_FIELDS)



import logging
from copy import deepcopy
from functools import partial

import pyproj
from shapely.geometry import box
from shapely.ops import cascaded_union, transform

import mapactionpy_controller.state_serialization as state_serialization
from mapactionpy_controller.recipe_layer import RecipeLayer

logger = logging.getLogger(__name__)


class RecipeFrame:
    """
    RecipeFrame - Includes an ordered list of layers for each Map Frame
    """
    OPTIONAL_FIELDS = ('extent', 'scale_text_element', 'spatial_ref_text_element')

    def __init__(self, frame_def, lyr_props, compatiblity_mode=0.3):
        # Required fields
        self.name = frame_def['name']
        # This is a list, but see note in `_parse_layers` method
        self.layers = self._parse_layers(frame_def['layers'], lyr_props)
        self.crs = self._parse_crs(frame_def, compatiblity_mode)

        # Optional fields
        self.extent = frame_def.get('extent', None)
        self.scale_text_element = frame_def.get('scale_text_element', None)
        self.spatial_ref_text_element = frame_def.get('spatial_ref_text_element', None)

    def _parse_layers(self, lyr_defs, lyr_props):
        # We create a seperate list and set here so that we can enforce unique layernames. However only
        # the list is returned. Client code is generally more readable and elegant if `self.layers` is a
        # list. This enforces that layer names must be unique in the json representation, however
        # theoretically allows client code to create multiple layers with identical names. The behaviour
        # in this circumstance is not known or tested and is entirely the client's responsiblity.
        recipe_lyrs_list = []
        lyrs_names_set = set()
        for lyr_def in lyr_defs:
            l_name = lyr_def['name']
            if l_name in lyrs_names_set:
                raise ValueError(
                    'Duplicate layer name {} in mapframe {}. Each layername within a'
                    ' mapframe must unique'.format(l_name, self.name))

            lyrs_names_set.add(l_name)
            recipe_lyrs_list.append(self._parse_single_layer(l_name, lyr_def, lyr_props))

        return recipe_lyrs_list

    def _parse_single_layer(self, l_name, lyr_def, lyr_props):
        # If lyr_def only includes as the only non-optional field is the `name` the retrive the layer
        # from the lyr_props object, and then apply the relevant optional properties
        # If not create a new recipe_lyr.
        present_manditatory_fields = set(lyr_def.keys()).difference(set(RecipeLayer.OPTIONAL_FIELDS))
        # opt_fields = [k for k in lyr_def.keys() if k in set(RecipeLayer.OPTIONAL_FIELDS)]
        if present_manditatory_fields == {'name'}:
            r_lyr = deepcopy(lyr_props.properties.get(l_name, l_name))
        else:
            r_lyr = RecipeLayer(lyr_def, lyr_props)

        # This only two fields that needs to be handled explictly at present are
        # `use_for_frame_extent` and `visable`
        try:
            r_lyr._apply_use_for_frame_extent(lyr_def)
        except AttributeError:
            pass

        if 'visible' in lyr_def:
            r_lyr.visible = lyr_def['visible']

        return r_lyr

    def _parse_crs(self, frame_def, compatiblity_mode):
        if compatiblity_mode >= 0.3:
            return frame_def['crs']

        # This hard coded values is for legacy recipes (schema==v0.2)
        return 'epsg:4326'

    def contains_layer(self, requested_layer_name):
        """
        Gets a layer by name.
        Returns a boolean
        """
        return requested_layer_name in [lyr.name for lyr in self.layers]

    def get_layer(self, requested_layer_name):
        """
        Gets a layer by name.
        Returns the RecipeLayer object
        Raises ValueError if the requested_layer_name does not exist
        """
        # We trust that the layer names are unique
        for lyr in self.layers:
            if lyr.name == requested_layer_name:
                return lyr

        raise ValueError(
            'The requested layer {} does not exist in the map frame {}'.format(
                requested_layer_name, self.name))

    def _filter_lyr_for_use_in_frame_extent(self):
        """
        The layer's `use_for_frame_extent` can have one of three values; True, False or None.
        * If one or more layers in a frame has `use_for_frame_extent==True` then those layers are used to
          determine the frame's extent (Whitelist).
        * If no layer has `use_for_frame_extent==True` and one or more layers in a frame has
          `use_for_frame_extent==False` Every layer except those will be used to determine the frame's
          extent (Blacklist).
        * If every layer has `use_for_frame_extent is None` then every layer will be used to determine the
          frame's extent (Default).
        """
        # White list
        extent_lyrs = [lyr for lyr in self.layers if lyr.use_for_frame_extent]

        # Black List
        if not extent_lyrs:
            # We already know that there are no cases where lyr.use_for_frame_extent==True
            extent_lyrs = [lyr for lyr in self.layers if lyr.use_for_frame_extent is None]

        # Default
        if not extent_lyrs:
            extent_lyrs = self.layers

        # Check that the extent has been defined for all the relevant layer.
        if not all([hasattr(lyr, 'extent') and hasattr(lyr, 'crs') for lyr in extent_lyrs]):
            raise ValueError('Cannot determine the layer extent for the relevant layers')

        return extent_lyrs

    def calc_extent(self, **kwargs):
        """
        The layer's `use_for_frame_extent` can have one of three values; True, False or None.
        * If one or more layers in a frame has `use_for_frame_extent==True` then those layers are used to
          determine the frame's extent (Whitelist).
        * If no layer has `use_for_frame_extent==True` and one or more layers in a frame has
          `use_for_frame_extent==False` Every layer except those will be used to determine the frame's
          extent (Blacklist).
        * If every layer has `use_for_frame_extent is None` then every layer will be used to determine the
          frame's extent (Default).
        """
        recipe = kwargs['state']

        # Convert all of the lyr.extents into the frame.crs
        projected_lyr_extents = []
        to_crs = pyproj.Proj(init=self.crs)
        # print('to_frame_crs = {}'.format(self.crs))

        for r_lyr in self._filter_lyr_for_use_in_frame_extent():
            # Get the projection transformation
            # print('from_lyr_crs = {}'.format(r_lyr.crs))

            if r_lyr.crs:
                project_func = partial(
                    pyproj.transform,
                    pyproj.Proj(init=r_lyr.crs),
                    to_crs
                )
                # print('from_lyr_crs = {}'.format(r_lyr.crs))

                # Create a shapely box from the lyr's bounds
                l_ext = box(*r_lyr.extent)
                # reproject the lyr bounds
                projected_lyr_extents.append(transform(project_func, l_ext))

        # print('projected_lyr_extents = {}'.format(projected_lyr_extents))
        # Now get the union of all of the extents
        self.extent = None
        if projected_lyr_extents:
            self.extent = cascaded_union(projected_lyr_extents).bounds

        return recipe

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        return not self.__eq__(other)

    def __getstate__(self):
        return state_serialization.get_state_optional_fields(self, RecipeFrame.OPTIONAL_FIELDS)

    def __setstate__(self, state):
        state_serialization.set_state_optional_fields(self, state, RecipeFrame.OPTIONAL_FIELDS)



from jsonschema import ValidationError
import os
from unittest import TestCase
import mapactionpy_controller.config_verify as config_verify


class TestConfigVerify(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.all_matching_cmf_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cmf_pointing_to_matching_lp_mcb_and_lyrs.json')
        self.mismatching_cmf_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cmf_pointing_to_mismatching_lp_mcb_and_lyrs.json')
        self.path_to_valid_cmf_des = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        self.path_to_invalid_cmf_des = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cmf_description_one_file_and_one_dir_not_valid.json')

    def test_check_cmf_description(self):
        # A valid cmf description
        cv = config_verify.ConfigVerifier(self.path_to_valid_cmf_des, ['.lyr'])
        try:
            cv.check_cmf_description()
            self.assertTrue(True)
        except ValueError as ve:
            self.fail(ve)

        # An Invalid cmf description
        cv = config_verify.ConfigVerifier(self.path_to_invalid_cmf_des, ['.lyr'])
        with self.assertRaises(ValueError):
            cv.check_cmf_description()

    def test_check_lyr_props_vs_rendering_dir(self):
        cv = config_verify.ConfigVerifier(self.mismatching_cmf_path, ['.lyr'])

        with self.assertRaises(ValueError):
            cv.check_lyr_props_vs_rendering_dir()

        cv = config_verify.ConfigVerifier(self.all_matching_cmf_path, ['.lyr'])
        try:
            cv.check_lyr_props_vs_rendering_dir()
            self.assertTrue(True)
        except ValueError as ve:
            self.fail(ve)

    def test_check_lyr_props_vs_map_cookbook(self):
        cv = config_verify.ConfigVerifier(self.mismatching_cmf_path, ['.lyr'])

        with self.assertRaises(ValueError):
            cv.check_lyr_props_vs_map_cookbook()

        cv = config_verify.ConfigVerifier(self.all_matching_cmf_path, ['.lyr'])
        try:
            cv.check_lyr_props_vs_map_cookbook()
            self.assertTrue(True)
        except ValueError as ve:
            self.fail(ve)

    def test_check_config_file_schemas(self):
        schema_errors_cmf_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'config_schemas',
            'fixture_cmf_pointing_schema_errors.json'
        )

        cv = config_verify.ConfigVerifier(schema_errors_cmf_path, ['.lyr'])
        with self.assertRaises(ValidationError):
            cv.check_json_file_schemas()

        cv = config_verify.ConfigVerifier(self.all_matching_cmf_path, ['.lyr'])
        cv.check_json_file_schemas()
        self.assertTrue(True)



# from mapactionpy_controller.steps import Step

# # Load event
# Step(
#     myfunc,
#     'Checking that the Humanitarian Event description file opens correctly',
#     'The Humanitarian Event description file opened correctly',
#     'Failed to open the Humanitarian Event description file correctly',
# ),

# # Load CMF Desc
# Step(
#     myfunc,
#     'Checking that the Crash Move Folder description file opens correctly',
#     'The Crash Move Folder description file opened correctly',
#     'Failed to open the Crash Move Folder description file correctly',
# ),

# # Load Recipe(s)
# Step(
#     myfunc,
#     'Checking that the Map Recipe file opens correctly',
#     'The Map Recipe file opened correctly',
#     'Failed to open the Map Recipe file correctly',
# ),

# # Combine recipe with layer_props
# Step(
#     myfunc,
#     'Combining Map Recipe with Layer Properties',
#     'Combined Map Recipe with Layer Properties',
#     'Failed to combined Map Recipe with Layer Properties',
# ),

# # Combine Recipe and event
# Step(
#     myfunc,
#     'Combining Map Recipe with Humanitarian Event',
#     'Combined Map Recipe with Humanitarian Event',
#     'Failed to combined Map Recipe with Humanitarian Event',
# ),

# # Check naming convention for all data in 2_Active_Data

# # Check that each file in the right sub folder within 2_Active_Data


# #  - update regexs
# Step(
#     myfunc,
#     'Updating data search regular expressions with Humanitarian Event details',
#     'Updated data search regular expressions with Humanitarian Event details',
#     'Failed to Update data search regular expressions with Humanitarian Event details'
# ),

# #  - search for data
# #  - check schema of data
# #  - check for multiple shapefile matches
# #  - calucate data-source credit text
# # Get Template
# # Get Oritenation
# # 'cook()'
# # - set zoom
# # alignLegend
# # Export
# #  - Create export dir
# #  - do_export jpeg, pdf, thumbnail
# #  - export_atlas (if required)
# #  - create zip file
# # Upload to MDR



import errno
import glob
import logging
import math
import os
from operator import itemgetter
import re
from shutil import copyfile
from zipfile import ZipFile

from slugify import slugify

from mapactionpy_controller.crash_move_folder import CrashMoveFolder


logger = logging.getLogger(__name__)

# abstract class
# Done using the "old-school" method described here, without using the abs module
# https://stackoverflow.com/a/25300153


class BaseRunnerPlugin(object):
    def __init__(self, hum_event, ** kwargs):
        self.hum_event = hum_event
        self.cmf = CrashMoveFolder(self.hum_event.cmf_descriptor_path)

        if not self.cmf.verify_paths():
            raise ValueError("Cannot find paths and directories referenced by cmf {}".format(self.cmf.path))

        if self.__class__ is BaseRunnerPlugin:
            raise NotImplementedError(
                'BaseRunnerPlugin is an abstract class and cannot be instantiated directly')

    def get_projectfile_extension(self, **kwargs):
        raise NotImplementedError(
            'BaseRunnerPlugin is an abstract class and the `get_projectfile_extension`'
            ' method cannot be called directly')

    def get_lyr_render_extension(self, **kwargs):
        raise NotImplementedError(
            'BaseRunnerPlugin is an abstract class and the `get_lyr_render_extension`'
            ' method cannot be called directly')

    def _get_all_templates_by_regex(self, recipe):
        """
        Gets the fully qualified filenames of map templates, which exist in `self.cmf.map_templates` whose
        filenames match the regex `recipe.template`.

        @param recipe: A MapRecipe object.
        @returns: A list of all of the templates, stored in `cmf.map_templates` whose
                 filename matches the regex `recipe.template` and that have the extention
                 `self.get_projectfile_extension()`
        """
        def _is_relevant_file(f):
            extension = os.path.splitext(f)[1]
            logger.debug('checking file "{}", with extension "{}", against pattern "{}" and "{}"'.format(
                f, extension, recipe.template, self.get_projectfile_extension()
            ))
            if re.search(recipe.template, f):
                logger.debug('file {} matched regex'.format(f))
                f_path = os.path.join(self.cmf.map_templates, f)
                logger.debug('file {} joined with self.cmf.map_templates "{}"'.format(f, f_path))
                return (os.path.isfile(f_path)) and (extension == self.get_projectfile_extension())
            else:
                return False

        # TODO: This results in calling `os.path.join` twice for certain files
        logger.debug('searching for map templates in; {}'.format(self.cmf.map_templates))
        all_filenames = os.listdir(self.cmf.map_templates)
        logger.debug('all available template files:\n\t{}'.format('\n\t'.join(all_filenames)))
        relevant_filenames = [os.path.realpath(os.path.join(self.cmf.map_templates, fi))
                              for fi in all_filenames if _is_relevant_file(fi)]
        logger.debug('possible template files:\n\t{}'.format('\n\t'.join(relevant_filenames)))
        return relevant_filenames

    def _get_template_by_aspect_ratio(self, template_aspect_ratios, target_ar):
        """
        Selects the template which best matches the required aspect ratio.

        @param possible_aspect_ratios: A list of tuples. For each tuple the first element is the path to the
                                       template. The second element is the relevant aspect ratio for that
                                       template. Typically this would be generated by
                                       `get_aspect_ratios_of_templates()`.
        @param target_ar: The taret aspect ratio - typically the aspect ratio of the bounding box for the country
                          being mapped.
        @returns: The path of the template with the best matching aspect ratio.
        """
        logger.info('Selecting from available templates based on the most best matching aspect ratio')

        # Target is more landscape than the most landscape template
        most_landscape = max(template_aspect_ratios, key=itemgetter(1))
        if most_landscape[1] < target_ar:
            logger.info('Target area of interest is more landscape than the most landscape template')
            return most_landscape[0]

        # Target is more portrait than the most portrait template
        most_portrait = min(template_aspect_ratios, key=itemgetter(1))
        if most_portrait[1] > target_ar:
            logger.info('Target area of interest is more portrait than the most portrait template')
            return most_portrait[0]

        # The option with the smallest aspect ratio that is larger than target_ar
        larger_ar = min(
            [(templ_path, templ_ar) for templ_path, templ_ar in template_aspect_ratios if templ_ar >= target_ar],
            key=itemgetter(1))
        # The option with the largest aspect ratio that is smaller than target_ar
        smaller_ar = max(
            [(templ_path, templ_ar) for templ_path, templ_ar in template_aspect_ratios if templ_ar <= target_ar],
            key=itemgetter(1))

        # Linear combination:
        # if (2*target_ar) > (larger_ar[1] + smaller_ar[1]):
        #     return larger_ar[0]

        # asmith: personally I think that this is the better option, but will go with the linear combination for now
        # logarithmic combination
        if (2*math.log(target_ar)) > (math.log(larger_ar[1]) + math.log(smaller_ar[1])):
            logger.info('Aspect ratio of the target area of interest lies between the aspect ratios of the'
                        ' available templates')
            return larger_ar[0]

        return smaller_ar[0]

    def get_aspect_ratios_of_templates(self, possible_templates, recipe):
        """
        Plugins are required to implement this method.

        The implementation should calculate the aspect ratio of the principal map frame within the list of
        templates. The definition of "principal" is left to the plugin, though is typically the largest map
        frame.

        @param possible_templates: A list of paths to possible templates
        @returns: A list of tuples. For each tuple the first element is the path to the template. The second
                  element is the aspect ratio of the largest* map frame within that template.
                  See `_get_largest_map_frame` for the description of how largest is determined.
        @raises NotImplementedError: In the base class.
        """
        raise NotImplementedError(
            'BaseRunnerPlugin is an abstract class and the `get_aspect_ratios_of_templates`'
            ' method cannot be called directly')

    def _get_aspect_ratio_of_bounds(self, bounds):
        minx, miny, maxx, maxy = bounds
        dx = (maxx - minx) % 360  # Accounts for the case where the bounds stradles the 180 meridian
        dy = maxy - miny

        return float(dx)/dy

    def get_templates(self, **kwargs):
        """
        Updates the recipe's `template_path` value. The result is the absolute path to the template.

        To select the appropriate template it uses two inputs.
        * The `recipe.template` value, which is a regex for the filename of the possible templates
        * The target asspect ratio. If the aspect ratio of the target data can be determined then this is
          also used to select the best matching template, amogst those which match the regex. If the
          target ratio cannot be determined fromsource gis data, then the target ratio of 1.0 will be
          used.
        """
        recipe = kwargs['state']
        # If there already is a valid `recipe.map_project_path` just skip with method
        if recipe.map_project_path:
            if os.path.exists(recipe.map_project_path):
                return recipe
            else:
                raise ValueError('Unable to locate map project file: {}'.format(recipe.map_project_path))

        # use `recipe.template` as regex to locate one or more templates
        possible_templates = self._get_all_templates_by_regex(recipe)

        # Select the template with the most appropriate aspect ratio
        possible_aspect_ratios = self.get_aspect_ratios_of_templates(possible_templates, recipe)

        mf = recipe.get_frame(recipe.principal_map_frame)
        # Default value
        target_aspect_ratio = 1.0
        # If the MapFrame's target extent is not None, then use that:
        if mf.extent:
            target_aspect_ratio = self._get_aspect_ratio_of_bounds(mf.extent)

        # use logic to workout which template has best aspect ratio
        # obviously not this logic though:
        recipe.template_path = self._get_template_by_aspect_ratio(possible_aspect_ratios, target_aspect_ratio)

        # TODO re-enable "Have the input files changed?"
        # Have the input shapefiles changed?
        return recipe

    # TODO: asmith 2020/03/03
    # 1) Please avoid hardcoding the naming convention for the mxds wherever possible. The Naming Convention
    # classes can avoid the need to hardcode the naming convention for the input mxd templates. It might be
    # possible to avoid the need to hardcode the naming convention for the output mxds using a
    # String.Template be specified within the Cookbook?
    # https://docs.python.org/2/library/string.html#formatspec
    # https://www.python.org/dev/peps/pep-3101/
    #
    # 2) This only checks the filename for the mxd - it doesn't check the values within the text element of
    # the map layout view (and hence the output metadata).
    def get_next_map_version_number(self, mapNumberDirectory, mapNumber, mapFileName):
        versionNumber = 0
        files = glob.glob(mapNumberDirectory + os.sep + mapNumber+'-v[0-9][0-9]-' + mapFileName + '.mxd')
        for file in files:
            versionNumber = int(os.path.basename(file).replace(mapNumber + '-v', '').replace(('-' + mapFileName+'.mxd'), ''))  # noqa
        versionNumber = versionNumber + 1
        return versionNumber

    # TODO Is it possible to aviod the need to hardcode the naming convention for the output mxds? Eg could a
    # String.Template be specified within the Cookbook?
    # https://docs.python.org/2/library/string.html#formatspec
    # https://www.python.org/dev/peps/pep-3101/
    def create_ouput_map_project(self, **kwargs):
        recipe = kwargs['state']
        # Create `mapNumberDirectory` for output
        output_dir = os.path.join(self.cmf.map_projects, recipe.mapnumber)

        if not(os.path.isdir(output_dir)):
            os.mkdir(output_dir)

        # Construct output MXD/QPRJ name
        logger.debug('About to create new map project file for product "{}"'.format(recipe.product))
        output_map_base = slugify(recipe.product)
        logger.debug('Set output name for new map project file to "{}"'.format(output_map_base))
        recipe.version_num = self.get_next_map_version_number(output_dir, recipe.mapnumber, output_map_base)
        output_map_name = '{}-v{}-{}{}'.format(
            recipe.mapnumber, str(recipe.version_num).zfill(2), output_map_base, self.get_projectfile_extension())
        recipe.map_project_path = os.path.abspath(os.path.join(output_dir, output_map_name))
        logger.debug('Path for new map project file; {}'.format(recipe.map_project_path))
        logger.debug('Map Version number; {}'.format(recipe.version_num))

        # Copy `src_template` to `recipe.map_project_path`
        copyfile(recipe.template_path, recipe.map_project_path)

        return recipe

    def export_maps(self, **kwargs):
        """
        Generates all file for export.

        Accumulate some of the parameters for export XML, then calls
        _do_export(....) to do that actual work
        """
        recipe = kwargs['state']
        export_params = {}
        properties = {}  # For properties from MapAction Toolbox
        try:
            export_params = self._create_export_dir(export_params, recipe)
            if 'properties' in kwargs:
                properties = kwargs['properties']
                for key in list(properties.keys()):
                    export_params[str(key)] = properties[str(key)]
            export_params = self._do_export(export_params, recipe)
        except Exception as exp:
            logger.error('Failed to export the map. export_params = "{}"'.format(export_params))
            logger.error(exp)

        self.zip_exported_files(export_params)

    def _create_export_dir(self, export_params, recipe):
        # Accumulate parameters for export XML
        version_str = "v" + str(recipe.version_num).zfill(2)
        export_directory = os.path.abspath(
            os.path.join(self.cmf.export_dir, recipe.mapnumber, version_str))
        export_params["exportDirectory"] = export_directory

        try:
            os.makedirs(export_directory)
        except OSError as exc:  # Python >2.5
            # Note 'errno.EEXIST' is not a typo. There should be two 'E's.
            # https://docs.python.org/2/library/errno.html#errno.EEXIST
            if exc.errno == errno.EEXIST and os.path.isdir(export_directory):
                pass
            else:
                raise

        return export_params

    def _do_export(self, export_params, recipe):
        raise NotImplementedError(
            'BaseRunnerPlugin is an abstract class and the `export_maps`'
            ' method cannot be called directly')

    def zip_exported_files(self, export_params):
        # Get key params as local variables
        core_file_name = export_params['coreFileName']
        export_dir = export_params['exportDirectory']
        mdr_xml_file_path = export_params['exportXmlFileLocation']
        jpg_path = export_params['jpgFileLocation']
        png_thumbnail_path = export_params['pngThumbNailFileLocation']
        # And now Zip
        zipFileName = core_file_name+".zip"
        zipFileLocation = os.path.join(export_dir, zipFileName)

        with ZipFile(zipFileLocation, 'w') as zipObj:
            zipObj.write(mdr_xml_file_path, os.path.basename(mdr_xml_file_path))
            zipObj.write(jpg_path, os.path.basename(jpg_path))
            zipObj.write(png_thumbnail_path, os.path.basename(png_thumbnail_path))
            if (len(export_params.get("emfFileLocation", "")) > 0):
                zipObj.write(export_params['emfFileLocation'], os.path.basename(export_params['emfFileLocation']))

            # TODO: asmith 2020/03/03
            # Given we are explictly setting the pdfFileName for each page within the DDPs
            # it is possible return a list of all of the filenames for all of the PDFs. Please
            # can we use this list to include in the zip file. There are edge cases where just
            # adding all of the pdfs in a particular directory might not behave correctly (eg if
            # the previous run had crashed midway for some reason)
            for pdf in os.listdir(export_dir):
                if pdf.endswith(".pdf"):
                    zipObj.write(os.path.join(export_dir, pdf),
                                 os.path.basename(os.path.join(export_dir, pdf)))
        print("Export complete to " + export_dir)

    def build_project_files(self, **kwargs):
        raise NotImplementedError(
            'BaseRunnerPlugin is an abstract class and the `build_project_files`'
            ' method cannot be called directly')



import subprocess
import sys
from setuptools import setup, find_packages
from os import path, environ

_base_version = '1.1.0'

root_dir = path.abspath(path.dirname(__file__))


def readme():
    with open(path.join(root_dir, 'README.md')) as f:
        return f.read()


# See https://packaging.python.org/guides/single-sourcing-package-version/
# This uses method 4 on this list combined with other methods.
def _get_version_number():
    travis_build = environ.get('TRAVIS_BUILD_NUMBER')
    travis_tag = environ.get('TRAVIS_TAG')
    repo_slug = environ.get('TRAVIS_REPO_SLUG')
    if repo_slug:
        repo_slug = repo_slug.lower().strip()

    if travis_build and (repo_slug == 'mapaction/mapactionpy_controller'):
        if travis_tag:
            version = travis_tag
        else:
            version = '{}.dev{}'.format(_base_version, travis_build)

        with open(path.join(root_dir, 'VERSION'), 'w') as version_file:
            version_file.write(version.strip())
    else:
        try:
            ver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'])
            version = '{}+local.{}'.format(_base_version, ver.decode('ascii').strip())
        except Exception:
            with open(path.join(root_dir, 'VERSION')) as version_file:
                version = version_file.read().strip()

    return version


def can_import_geo_packages():
    try:
        # Test PyProj
        import pyproj  # noqa: F401

        # Test Shapely
        from shapely.geometry import box  # noqa: F401

        # Test GDAL
        from osgeo import gdal  # noqa: F401
        from osgeo import ogr  # noqa: F401
        from osgeo import osr  # noqa: F401
        from osgeo import gdal_array  # noqa: F401
        from osgeo import gdalconst  # noqa: F401

        # Test Fiona
        import fiona  # noqa: F401

        # Test RTree
        from rtree import index  # noqa: F401

        # Test geopandas
        import geopandas  # noqa: F401

        return True
    except ImportError:
        return False


def _get_requires_list():
    # Items for which the version does not need to be pinned to support py2.7
    requires = [
        'chevron',
        'humanfriendly',
        'jsonpickle',
        'jsonschema',
        'requests',
        'pyyaml',
        'pyshp',
        'six',
        'slugify'
    ]

    if (sys.version_info.major == 2):
        # Items for which do need a pinned version to support py2.7
        # For debate whether this is useful 'pyreproj==1.0.1'
        requires.extend([
            'pyrsistent<=0.16.1',
            'pycountry<=19.8.18'
        ])
    else:
        requires.extend([
            'pyrsistent',
            'pycountry',
        ])

    if sys.platform == 'win32' and not can_import_geo_packages():
        requires.extend([
            'pytz',
            'mapactionpy_controller_dependencies'
        ])
    else:
        try:
            # Test the underlying version of GDAL, so that we can install the matching python bindings
            gdal_cmd_ver = subprocess.check_output(['gdal-config', '--version'])
            gdal_ver = gdal_cmd_ver.decode('ascii').strip()
            gdal_str = 'GDAL=={}'.format(gdal_ver)
        except OSError:
            gdal_str = 'GDAL'

        requires.extend([
            'Fiona',
            'pyproj',
            'Shapely',
            gdal_str,
            'Rtree',
            'geopandas'
        ])

    return requires


setup(
    name='mapactionpy_controller',
    version=_get_version_number(),
    description='Controls the workflow of map and infographic production',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='http://github.com/mapaction/mapactionpy_controller',
    author='MapAction',
    author_email='github@mapaction.com',
    license='GPL3',
    entry_points={
        'console_scripts': [
            'mapchef=mapactionpy_controller.cli:entry_point'
        ]
    },
    packages=find_packages(),
    package_dir={'mapactionpy_controller': 'mapactionpy_controller'},
    package_data={'mapactionpy_controller': ['schemas/*.schema']},
    include_package_data=True,
    install_requires=_get_requires_list(),
    test_suite='unittest',
    tests_require=['unittest'],
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"
    ])



from mapactionpy_controller import _get_validator_for_config_schema

validate_against_atlas_schema = _get_validator_for_config_schema('atlas-v0.2.schema')


class RecipeAtlas:
    def __init__(self, atlas_def, recipe, lyr_props):
        validate_against_atlas_schema(atlas_def)

        # Required fields
        self.map_frame = atlas_def["map_frame"]
        self.layer_name = atlas_def["layer_name"]
        self.column_name = atlas_def["column_name"]

        # Compare the atlas definition with the other parts of the recipe definition
        if recipe.contains_frame(self.map_frame):
            m_frame = recipe.get_frame(self.map_frame)
        else:
            raise ValueError(
                'The Map Recipe definition is invalid. The "atlas" section refers to a map_frame '
                ' ({}) that does not exist in the "map_frames" section of the recipe.'.format(
                    self.map_frame)
            )

        if m_frame.contains_layer(self.layer_name):
            lyr = m_frame.get_layer(self.layer_name)
        else:
            raise ValueError(
                'The Map Recipe definition is invalid. The "atlas" section refers to a layer_name '
                ' ({}) that does not exist in the relevant "map_frame" ({}) section of the recipe.'
                ''.format(self.layer_name, self.map_frame)
            )

        if self.column_name not in lyr.data_schema['required']:
            raise ValueError(
                'The Map Recipe definition is invalid. The "atlas" section refers to a column_name '
                ' ({}) that does not exist in the schema of the relevant layer ({}).'
                ''.format(self.column_name, lyr.name)
            )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        return not self.__eq__(other)



import argparse
import os

import mapactionpy_controller.check_naming_convention as cnc
import mapactionpy_controller.config_verify as config_verify
import mapactionpy_controller.plugin_controller as plugin_controller
from mapactionpy_controller.main_stack import process_stack

VERB_BUILD = 'build'
VERB_CREATE = 'create'
VERB_LIST = 'list'
VERB_UPDATE = 'update'
VERB_UPLOAD = 'upload'
VERB_VERIFY = 'verify'


def noun_defaultcmf_print_output(args):
    if args.verb == VERB_VERIFY:
        cv_steps = config_verify.get_config_verify_steps(args.cmf_desc_path, ['.lyr'])
        # process_stack(cv_steps, None)
        cv_steps.extend(cnc.get_defaultcmf_step_list(args.cmf_desc_path))
        process_stack(cv_steps, None)
    else:
        raise NotImplementedError(args)


def noun_humevent_print_output(args):
    raise NotImplementedError(args)


def noun_gisdata_print_output(args):
    if args.verb == VERB_VERIFY:
        nc_steps = cnc.get_active_data_step_list(args.humevent_desc_path)
        process_stack(nc_steps, None)
        # print(nc_steps)
    else:
        raise NotImplementedError(args)


def noun_maps_print_output(args):
    if args.verb == VERB_BUILD:
        build_maps(args.humevent_desc_path, args.map_number, args.dry_run)
    else:
        raise NotImplementedError(args)


def build_maps(humevent_desc_path, map_number, dry_run):
    # build_steps = config_verify.get_config_verify_steps(args.cmf_desc_path, ['.lyr'])
    # build_steps.append(cnc.get_defaultcmf_step_list(args.cmf_desc_path, False))
    # build_steps.append(cnc.get_active_data_step_list(args.humevent_desc_path, True))
    # main_stack.process_stack(build_steps)
    my_runner = process_stack(plugin_controller.get_plugin_step(), humevent_desc_path)
    process_stack(plugin_controller.get_cookbook_steps(my_runner, map_number, dry_run), None)

    # map_nums = None
    # if map_number:
    #     map_nums = [map_number]

    # for recipe in plugin_controller.select_recipes(my_cookbook, map_nums):
    #     product_steps = plugin_controller.get_per_product_steps(my_runner, recipe.mapnumber, recipe.product)
    #     process_stack(product_steps, recipe)


all_nouns = {
    'defaultcmf': ('Print info about the Default Crash Move Folder', noun_defaultcmf_print_output),
    'humevent': ('Access and update info about a Humanitarian Event', noun_humevent_print_output),
    'gisdata': ('Access and update info about the GIS data collected for a Humanitarian Event',
                noun_gisdata_print_output),
    'maps': ('Access, build, update and upload maps for a Humanitarian Event', noun_maps_print_output)
}

all_verbs = {
    'defaultcmf': [
        (VERB_VERIFY, 'Verify the internal self consistency of the Default Crash Move Folder, without reference'
                      ' to any country or situational GIS data')
    ],
    'humevent': [
        (VERB_CREATE, 'Create a new Humanitarian Event description file. Values are read from relevant'
                      ' environment variables'),
        (VERB_UPDATE, 'Update an existing Humanitarian Event description file. Values are read from relevant'
                      ' environment variables'),
        (VERB_VERIFY, 'Verify the internal self consistency of the Humanitarian Event description, without'
                      ' reference to any country or situational GIS data')
    ],
    'gisdata': [
        (VERB_VERIFY, 'Verify that the "active" GIS data adheres to relevant GIS standards (eg Data Naming'
                      ' Convention and Schemas)')
    ],
    'maps': [
        (VERB_BUILD, 'Create the maps described in the MapCookbook. Existing maps are recreated if any'
                     ' of the inputs have changed.'),
        (VERB_UPLOAD, 'Upload maps to the publishing site')
    ]
}


def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error('The file "%s" does not exist!' % arg)
    else:
        return arg


def _add_verbs_to_parser(verb_desc, parser):
    verb_grp = parser.add_mutually_exclusive_group(required=False)

    for verb, help_desc in verb_desc:
        flag = '--{}'.format(verb)

        verb_grp.add_argument(
            flag,
            action='store_const',
            dest='verb',
            help=help_desc,
            const=verb
        )


def _create_noun_parser(noun_str, subparser):
    noun_desc, noun_func = all_nouns[noun_str]

    parser = subparser.add_parser(noun_str)
    parser.description = (noun_desc)
    parser.set_defaults(func=noun_func)
    _add_verbs_to_parser(all_verbs[noun_str], parser)
    return parser


def get_args():
    mainparser = argparse.ArgumentParser(
        prog='mapchef',
        description=(
            "This tool does everything you'd ever what to make maps automatically."
            " Please look at the sub commands for more details"
        )
    )

    prs_nouns = mainparser.add_subparsers(title='available subcommands',
                                          description=(
                                              'To perform any useful tasks you will need to use one of'
                                              ' the subcommands listed below. For more detailed information'
                                              ' try `{} < subcommand > help`'.format(mainparser.prog)),
                                          help=None)

    # Noun: defaultcmf
    prs_defaultcmf = _create_noun_parser('defaultcmf', prs_nouns)
    prs_defaultcmf.add_argument(
        'cmf_desc_path',
        metavar='crash-move-folder-description-file',
        help='The path to Crash Move Foler (CMF) description file.',
        type=lambda x: is_valid_file(prs_defaultcmf, x)
    )

    # Noun: humevent
    prs_humevent = _create_noun_parser('humevent', prs_nouns)
    # It is acceptable that the `humevent_desc_path` does not exist for `humevent create`
    prs_humevent.add_argument(
        'humevent_desc_path',
        metavar='humanitarian-event-description-file',
        help='The path to Humanitarian Event description file.'
    )

    # Noun: gisdata
    prs_gisdata = _create_noun_parser('gisdata', prs_nouns)
    prs_gisdata.add_argument(
        'humevent_desc_path',
        metavar='humanitarian-event-description-file',
        help='The path to Humanitarian Event description file.',
        type=lambda x: is_valid_file(prs_humevent, x)
    )

    # Noun: maps
    prs_maps = _create_noun_parser('maps', prs_nouns)
    prs_maps.add_argument(
        'humevent_desc_path',
        metavar='humanitarian-event-description-file',
        help='The path to Humanitarian Event description file.',
        type=lambda x: is_valid_file(prs_humevent, x)
    )

    prs_maps.add_argument(
        '--map-number',
        metavar='"Map Number"',
        help=('The number of an individual map to produce (eg "MA0123"). This must match a map'
              ' number name that exists in the MapCookbook. If this option is not specified then'
              ' all maps in the MapCookbook will be created.')
    )

    maps_options_grp = prs_maps.add_mutually_exclusive_group(required=False)

    maps_options_grp.add_argument(
        '--force',
        action='store_true',
        help=('Generate a new version of the specified products, even if no change is detected'
              ' in any of the input files. (`--force` and `--dry-run`  cannot'
              ' be specified together.)')
    )

    maps_options_grp.add_argument(
        '--dry-run',
        action='store_true',
        help=('Do not generate a new version of the specified products. Run through each step'
              ' and attempt to identify any potential errors. (`--dry-run` and `--force` cannot'
              ' be specified together.)')
    )

    return mainparser


def entry_point():
    mainparser = get_args()
    args = mainparser.parse_args()
    try:
        args.func(args)
    except AttributeError:
        mainparser.print_usage()


if __name__ == "__main__":
    entry_point()



from xml.etree.ElementTree import Element, SubElement, tostring
import xml.dom.minidom


class MapDoc:
    # Required to generate 'mapdoc' XML document
    # Needs mapdata in its own class so that it's held in the inner mapdata block:
    #
    # <?xml version="1.0" ?>
    # <mapdoc>
    #    <mapdata>

    def __init__(self, mapdata):
        self.mapdata = mapdata

    def to_xml(self):
        mapDoc = Element('mapdoc')

        mapdata = SubElement(mapDoc, 'mapdata')

        versionNumber = SubElement(mapdata, 'versionNumber')
        versionNumber.text = str(self.mapdata.versionNumber)

        mapNumber = SubElement(mapdata, 'mapNumber')
        mapNumber.text = str(self.mapdata.mapNumber)

        operationID = SubElement(mapdata, 'operationID')
        operationID.text = str(self.mapdata.operationID)

        sourceOrg = SubElement(mapdata, 'sourceorg')
        sourceOrg.text = str(self.mapdata.sourceorg)

        title = SubElement(mapdata, 'title')
        title.text = str(self.mapdata.title)

        ref = SubElement(mapdata, 'ref')
        ref.text = str(self.mapdata.ref)

        language = SubElement(mapdata, 'language')
        language.text = str(self.mapdata.language)

        countries = SubElement(mapdata, 'countries')
        countries.text = str(self.mapdata.countries)

        createdate = SubElement(mapdata, 'createdate')
        createdate.text = str(self.mapdata.createdate)

        createtime = SubElement(mapdata, 'createtime')
        createtime.text = str(self.mapdata.createtime)

        status = SubElement(mapdata, 'status')
        status.text = str(self.mapdata.status)

        xmin = SubElement(mapdata, 'xmin')
        xmin.text = str(self.mapdata.xmin)

        ymin = SubElement(mapdata, 'ymin')
        ymin.text = str(self.mapdata.ymin)

        xmax = SubElement(mapdata, 'xmax')
        xmax.text = str(self.mapdata.xmax)

        ymax = SubElement(mapdata, 'ymax')
        ymax.text = str(self.mapdata.ymax)

        proj = SubElement(mapdata, 'proj')
        proj.text = str(self.mapdata.proj)

        datum = SubElement(mapdata, 'datum')
        datum.text = str(self.mapdata.datum)

        qclevel = SubElement(mapdata, 'qclevel')
        qclevel.text = str(self.mapdata.qclevel)

        qcname = SubElement(mapdata, 'qcname')
        qcname.text = str(self.mapdata.qcname)

        access = SubElement(mapdata, 'access')
        access.text = str(self.mapdata.access)

        glideno = SubElement(mapdata, 'glideno')
        glideno.text = str(self.mapdata.glideno)

        summary = SubElement(mapdata, 'summary')
        summary.text = str(self.mapdata.summary)

        imagerydate = SubElement(mapdata, 'imagerydate')
        imagerydate.text = str(self.mapdata.imagerydate)

        datasource = SubElement(mapdata, 'datasource')
        datasource.text = str(self.mapdata.datasource)

        location = SubElement(mapdata, 'location')
        location.text = str(self.mapdata.location)

        themes = SubElement(mapdata, 'themes')
        for setTheme in self.mapdata.themes:
            theme = SubElement(themes, 'theme')
            theme.text = setTheme

        scale = SubElement(mapdata, 'scale')
        scale.text = str(self.mapdata.scale)

        papersize = SubElement(mapdata, 'papersize')
        papersize.text = str(self.mapdata.papersize)

        jpgresolutiondpi = SubElement(mapdata, 'jpgresolutiondpi')
        jpgresolutiondpi.text = str(self.mapdata.jpgresolutiondpi)

        pdfresolutiondpi = SubElement(mapdata, 'pdfresolutiondpi')
        pdfresolutiondpi.text = str(self.mapdata.pdfresolutiondpi)

        kmlresolutiondpi = SubElement(mapdata, 'kmlresolutiondpi')
        kmlresolutiondpi.text = str(self.mapdata.kmlresolutiondpi)

        mxdfilename = SubElement(mapdata, 'mxdfilename')
        mxdfilename.text = str(self.mapdata.mxdfilename)

        paperxmax = SubElement(mapdata, 'paperxmax')
        paperxmax.text = str(self.mapdata.paperxmax)

        paperxmin = SubElement(mapdata, 'paperxmin')
        paperxmin.text = str(self.mapdata.paperxmin)

        paperymax = SubElement(mapdata, 'paperymax')
        paperymax.text = str(self.mapdata.paperymax)

        paperymin = SubElement(mapdata, 'paperymin')
        paperymin.text = str(self.mapdata.paperymin)

        accessnotes = SubElement(mapdata, 'accessnotes')
        accessnotes.text = str(self.mapdata.accessnotes)

        product_type = SubElement(mapdata, 'product-type')
        product_type.text = str(self.mapdata.product_type)

        language_iso2 = SubElement(mapdata, 'language-iso2')
        language_iso2.text = str(self.mapdata.language_iso2)

        jpgfilename = SubElement(mapdata, 'jpgfilename')
        jpgfilename.text = str(self.mapdata.jpgfilename)

        pdffilename = SubElement(mapdata, 'pdffilename')
        pdffilename.text = str(self.mapdata.pdffilename)

        emffilename = SubElement(mapdata, 'emffilename')
        emffilename.text = self.mapdata.emffilename

        jpgfilesize = SubElement(mapdata, 'jpgfilesize')
        jpgfilesize.text = str(self.mapdata.jpgfilesize)

        pdffilesize = SubElement(mapdata, 'pdffilesize')
        pdffilesize.text = str(self.mapdata.pdffilesize)

        emffilesize = SubElement(mapdata, 'emffilesize')
        if self.mapdata.emffilesize is not None:
            emffilesize.text = str(self.mapdata.emffilesize)
        else:
            emffilesize.text = ""

        dom = xml.dom.minidom.parseString(tostring(mapDoc, encoding='utf-8', method='xml'))
        return(dom.toprettyxml())



# -*- coding: utf-8 -*-
import logging
import netrc
import os
from datetime import datetime

import pytz
from jira import JIRA

from mapactionpy_controller.task_renderer import TaskReferralBase

logger = logging.getLogger(__name__)


def _get_secrets_from_netrc():
    possible_netrc_locations = [os.path.join(os.environ[envar], '.netrc')
                                for envar in ['HOME', 'USERPROFILE'] if envar in os.environ]

    if 'MAPCHEF_NETRC' in os.environ:
        possible_netrc_locations.append(os.environ['MAPCHEF_NETRC'])

    secrets = None
    for netrc_path in possible_netrc_locations:
        try:
            secrets = netrc.netrc(netrc_path)
        except IOError:
            pass

    return secrets


def _check_jira_con(jira_con, username):
    # Check that the user is actually authenticated
    if not jira_con.myself()['emailAddress'] == username:
        err_msg = 'JIRA Connection. Unable to authenticate with JIRA. Please check details in `.netrc` file.'
        logger.error(err_msg)
        raise ValueError(err_msg)


class JiraClient():
    jira_con = None
    board_details = None

    def __init__(self):
        self._get_jira_board_details()
        secrets = _get_secrets_from_netrc()

        if not secrets:
            err_msg = 'Unable to locate or load suitable `.netrc` file for JIRA integration'
            logger.error(err_msg)
            raise ValueError(err_msg)

        try:
            username, account, apikey = secrets.authenticators(self.jira_hostname)
        except TypeError:
            err_msg = 'JIRA Connection. Unable to find details for machine "{}" `.netrc` file.'.format(
                self.jira_hostname)
            logger.error(err_msg)
            raise ValueError(err_msg)

        self.jira_con = JIRA(options={'server': account}, basic_auth=(username, apikey))
        _check_jira_con(self.jira_con, username)
        logger.debug('JIRA Connection. Details = {}'.format(self.jira_con.myself()))

    def _get_jira_board_details(self):
        # TODO read these in from a config file
        self.jira_hostname = 'mapaction.atlassian.net'
        self.project_key = 'PIPET'
        # The target column should be were the column where new issues are created
        self.target_column = '10110'
        self.common_task_fields = {
            'project': self.project_key,
            'issuetype': {'id': '10235'}
        }

    def __del__(self):
        try:
            self.jira_con.kill_session()
        except (TypeError, AttributeError):
            pass

    def task_handler(self, fail_threshold, msg, task_referal=None):
        logger.debug('JiraClient.task_handler called with status="{}", and msg="{}"'.format(
            fail_threshold, msg))

        assured_referal = self.ensure_task_referal_type(task_referal, msg, fail_threshold)

        if not assured_referal:
            logger.debug('JiraClient.task_handler; `None` value passed for task_referal parameter. Nothing to handle.')
            return

        unique_summary = assured_referal.get_task_unique_summary()
        task_desc = assured_referal.get_task_description()
        op_id = assured_referal.get_operation_id()

        j_issue = self.search_issue_by_unique_summary(unique_summary, op_id)

        if j_issue:
            # Update existing card and maybe move it back into "Doing" column
            self.update_jira_issue(j_issue, task_desc, fail_threshold)
        else:
            if fail_threshold > logging.INFO:
                # Create a new task
                self.create_new_jira_issue(unique_summary, task_desc, op_id)

    def ensure_task_referal_type(self, task_referal, msg, fail_threshold):
        """
        Check whether or not the `task_referal` is an instance of TaskReferralBase object. If it is the object is
        then it is returned unchanged. If not then an generic TaskReferralBase will be created and returned. The
        value of `str(task_referal)` will be used.

        @param task_referal: An object that may or may not be a TaskReferralBase object.
        @returns: If the `task_referal` param is an instance of TaskReferralBase object, then `task_referal` is
                  returned.
                  If `task_referal` param is NOT an instance of TaskReferralBase AND fail_threshold is logging.ERROR
                  then a new TaskReferralBase object is created (using `msg` and `str(task_referal)` for context).
                  Else `None` is returned.
        """
        if isinstance(task_referal, TaskReferralBase):
            logger.debug('JiraClient.ensure_task_referal_type found a TaskReferralBase object')
            return task_referal

        if task_referal and (fail_threshold > logging.WARNING):
            logger.debug('JiraClient.ensure_task_referal_type created a new TaskReferralBase object')
            return TaskReferralBase(None, msg=msg, other=str(task_referal))

        logger.debug('JiraClient.ensure_task_referal_type passed "{}" but returned `None`'.format(
            str(task_referal)
        ))
        return None

    def search_issue_by_unique_summary(self, search_summary, op_id):
        # Default if `op_id` is None
        jql_op_id = 'operational_id is EMPTY'
        if op_id:
            jql_op_id = 'operational_id ~ "{}"'.format(op_id)

        jql_str = 'project={} AND {} AND summary ~ "{}"'.format(self.project_key, jql_op_id, search_summary)
        found_issues = self.jira_con.search_issues(jql_str, maxResults=2)

        if found_issues:
            if len(found_issues) > 1:
                raise ValueError(
                    'More than one JIRA Issue found with the summary "{}". This suggests that additional'
                    ' issues have been raised manualy on the board "{}". Please ensure that there is exactly'
                    ' one issues with this summary, by deleting those which have not been created by the'
                    ' user "{}"'.format(
                        search_summary,
                        self.project_key,
                        self.jira_con.myself()['emailAddress']
                    )
                )
            else:
                return found_issues[0]
        else:
            return None

    def create_new_jira_issue(self, unique_summary, task_desc, op_id):
        flds = self.common_task_fields.copy()
        flds['summary'] = unique_summary
        flds['description'] = task_desc
        # This is the JIRA API's field ID for operational_id. To work this out execute:
        # ```
        # a = j.jira_con.createmeta(projectKeys=['PIPET'], issuetypeIds=[10235], expand='projects.issuetypes.fields')
        # print(a)
        # ```
        # Then search the output for your custom field name. Doubtless there is a programmatic way to do this.
        flds['customfield_10234'] = op_id

        new_task = self.jira_con.create_issue(fields=flds)
        # new_task.update(fields={'operational_id':op_id})
        # new_task.update(operational_id=op_id)
        print(new_task)
        # print('desc', new_task.fields.description)
        # print('opid', new_task.fields.operational_id)
        # for f in new_task.fields:
        #     print('field itr', f)

    def update_jira_issue(self, j_issue, task_desc, fail_threshold):
        now_utc = pytz.utc.localize(datetime.now())
        time_stamp = now_utc.strftime('%Y-%m-%d %H:%M:%S %Z%z')

        # prev_desc =
        if task_desc != j_issue.fields.description:
            j_issue.update(description=task_desc)

        if fail_threshold > logging.INFO:
            self.jira_con.add_comment(
                j_issue.id,
                'This Issue was still current when MapChef was run at {}'.format(time_stamp))

        # if (fail_threshold <= logging.INFO)   eg *is not* an warning or an error
        # and (j_issue.fields.status.id == self.target_column):  eg and still in the todo column
        # and (j_issue.fields.status.id == self.target_column):  eg and still in the todo column
        # then add comment (and possibly automatically close).
        # self.jira_con.add_comment(
        #     j_issue.id,
        #     'This Issue appeared to be resolved when MapChef was run at {}. Please manually'
        #     ' check that the outputs are as expected and then close this Issue.'.format(
        #         time_stamp))

        # if (fail_threshold > logging.INFO)   eg *IS* an warning or an error
        # and (not j_issue.fields.status.id == self.target_column):  # Not in the todo column
        # then it may be necessary to re-open the ticket.
        # self.jira_con.transition_issue(
        #     j_issue.id,
        #     comment='This Issue was still current on MapChef run at {}.'
        #     ' Moving the issue to the DOING column'.format(time_stamp))
        # )



import logging
from mapactionpy_controller.map_cookbook import MapCookbook
from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.steps import Step


class ConfigVerifier():
    def __init__(self, cmf_desc, lyr_file_exn_list):
        self.cmf_desc_path = cmf_desc
        self.lyr_file_exn_list = lyr_file_exn_list

    def check_cmf_description(self, **kwargs):
        self.cmf = CrashMoveFolder(self.cmf_desc_path)
        return ('The Crash Move Folder description file open correctly:\n"{}"\n'.format(
            self.cmf_desc_path
        ))

    def check_json_file_schemas(self, **kwargs):
        try:
            # JSON schema validation is implicit in the creation of these objects
            self.check_cmf_description()
            lp = LayerProperties(self.cmf, '', verify_on_creation=False)
            MapCookbook(self.cmf, lp, verify_on_creation=False)
            return('No json validation problems were detected in the parsing of these two'
                   ' files:\n"{}"\n"{}"'.format(lp.cmf.layer_properties, self.cmf.map_definitions)
                   )
        except ValueError:
            raise

    def check_lyr_props_vs_rendering_dir(self, **kwargs):
        cmf = CrashMoveFolder(self.cmf_desc_path)
        for lyr_exn in self.lyr_file_exn_list:
            try:
                LayerProperties(cmf, lyr_exn, verify_on_creation=True)
                return('No inconsistency detected between:\n'
                       ' * the contents of the layer properties json file:\n\t{props}\n'
                       ' * and layer rendering dir:\n\t{render}\n'.format(
                           props=cmf.layer_properties,
                           render=cmf.layer_rendering
                       ))
            except ValueError:
                raise

    def check_lyr_props_vs_map_cookbook(self, **kwargs):
        try:
            cmf = CrashMoveFolder(self.cmf_desc_path)
            lyrs = LayerProperties(cmf, '', verify_on_creation=False)
            MapCookbook(cmf, lyrs, verify_on_creation=True)
            return('No inconsistency detected between:\n'
                   ' * the contents of the layer properties json file:\n\t{props}\n'
                   ' * and the contents of the MapCookbook json:\n\t{cbook}\n'.format(
                       props=cmf.layer_properties,
                       cbook=cmf.map_definitions
                   ))
        except ValueError:
            raise


def get_config_verify_steps(cmf_desc_path, lyr_file_exn_list):
    cv = ConfigVerifier(cmf_desc_path, lyr_file_exn_list)

    config_verify_steps = [
        Step(
            cv.check_cmf_description,
            logging.WARNING,
            'Checking that the Crash Move Folder description file opens correctly',
            'The Crash Move Folder description file opened correctly',
            'Failed to open the Crash Move Folder description file correctly',
        ),
        Step(
            cv.check_json_file_schemas,
            logging.WARNING,
            'Checking that each of the configuration files matches their relevant schemas',
            'Each of the configuration files adheres to their relevant schemas',
            'Failed to verify one or more of the configuration files against the relevant schema',
        ),
        Step(
            cv.check_lyr_props_vs_rendering_dir,
            logging.WARNING,
            'Comparing the contents of the layer properties json file and the layer rendering directory',
            'Compared the contents of the layer properties json file and the layer rendering directory',
            'Inconsistancy found in between the contents of the layer properties json file and the layer'
            ' rendering directory'
        ),
        Step(
            cv.check_lyr_props_vs_map_cookbook,
            logging.WARNING,
            'Comparing the contents of the layer properties json file and the MapCookbook',
            'Compared the contents of the layer properties json file and the MapCookbook',
            'Inconsistancy found in between the contents of the layer properties json file and the MapCookbook'
        )
    ]

    return config_verify_steps



import logging
import traceback


class Step():
    """
    This represents the atomic "unit of work" within the map production automation.

    The contract for functions
        'state' - An object of any kind. Either generated by the previous Step or passed as the `initial_state`
                  parameter of `main_stack.process_stack` function.

    Return values:
        Any object. Does not need to be a dict. The return value will be wrapped in the dict prior to being
        passed to the next step.  kwargs = {'state' : return_value}

    Exceptions:

            pass_back['result'] = result
            pass_back['exp'] = exp
            pass_back['stack_trace'] = traceback.format_exc()


    :param func: A callable which will do the relevant "unit of work".
        * Must accept a **kwargs param. If there is a state object this is passed as `kwargs['state']`.
          The stack does not guarentee the type of object that will be passed as the state.
        * If the function completes successfully it should return the updated state object. This should be
          "bare" (ie not wrapped in a dict ala kwargs)
        * If the function does not complete successfully it should raise an exception.
    :param fail_threshold: Expresses the severity with which an exception from `func` should be treated. In
        any case the exception will be handled and a JIRA tasks logged as appropriate.
        * If = `logging.ERROR` - the exception will terminate the program.
        * If = `logging.WARNING` - the exception will not result in termination. A JRIA task will be logged
          the program will attempt to continue, though the results may not be want the user intended.
    :param running_msg: A descriptive string to display (ie on the terminal) whilst this step is running.
    :param complete_msg: A descriptive string to display (eg on the terminal or in the logs) when the step as
                         completed successfully.
    :param fail_msg: A descriptive string to display (eg on the terminal or in the logs) if the step fails to
                     complete successfully.
    :returns: Either:
        * An object which will be passed to the next Step. The returned value will be wrapped in the dict prior
          to being passed to the next step.  kwargs = {'state' : return_value}
        * Or one or more Step objects. If multiple Step objects then these should be contained within a list.
          Any Step objects will be added to the stack in reverse order. (eg the list object in the list will be
          the next to be executed).
    """

    def __init__(self, func, fail_threshold, running_msg, complete_msg, fail_msg):
        self.func = func
        self.fail_threshold = fail_threshold
        self.running_msg = running_msg
        self.complete_msg = complete_msg
        self.fail_msg = fail_msg

    def run(self, set_feedback, **kwargs):
        pass_back = kwargs.copy()

        try:
            result = self.func(**kwargs)
            # long_msg = '{}\n{}'.format(self.complete_msg, result)
            pass_back['result'] = result
            set_feedback(logging.INFO, self.complete_msg, self, **pass_back)
            return result
        # Because this is defacto part of the main stack we do want to catch
        # a top level Exception here:
        except Exception as exp:
            pass_back['exp'] = exp
            pass_back['stack_trace'] = traceback.format_exc()
            set_feedback(self.fail_threshold, self.fail_msg, self, **pass_back)
            # set_feedback(logging.ERROR, self.fail_msg, self, **pass_back)

            # Do we want to raise an ERROR or a WARNING?
            if self.fail_threshold >= logging.ERROR:
                # print(kwargs.get('state', None))
                raise exp

            # If this is just a warning then we return the unaltered state object
            # TODO Review is it is possible in the case of warnings, to pass back an updated
            # state as an args to the relevant Exception.
            return kwargs.get('state', None)




from unittest import TestCase
import fixtures
import jsonschema
import os
import six
import yaml

from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.map_recipe import MapRecipe
import mapactionpy_controller.recipe_layer as recipe_layer

try:
    from unittest import mock
except ImportError:
    import mock


class TestRecipeLayer(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.path_to_valid_cmf_des = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        self.path_to_invalid_cmf_des = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cmf_description_one_file_and_one_dir_not_valid.json')

        self.cmf = CrashMoveFolder(self.path_to_valid_cmf_des, verify_on_creation=False)
        self.lyr_props = LayerProperties(self.cmf, '', verify_on_creation=False)

    def test_layer_data_schema(self):

        null_schema = True

        passing_schema = yaml.safe_load(r"""
required:
    - name_en
properties:
    geometry_type:
        items:
            enum:
                - MultiPolygon
                - Polygon
        additionalItems: false
    crs:
        items:
            enum:
                - EPSG:2090
        additionalItems: false
""")

        # Note the missing `enum` blocks
        failing_schema = yaml.safe_load(r"""
required:
    - name_en
properties:
    geometry_type:
        items:
            - MultiPolygon
            - Polygon
        additionalItems: false
    crs:
        items:
            - EPSG:2090
        additionalItems: false
""")

        cmf = CrashMoveFolder(
            os.path.join(self.parent_dir, 'example', 'cmf_description_relative_paths_test.json'))
        cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'cookbooks', 'fixture_layer_properties_for_atlas.json'
        )

        # Two cases where data schema is valid yaml
        for test_schema in [null_schema, passing_schema]:
            with mock.patch('mapactionpy_controller.data_schemas.yaml.safe_load') as mock_safe_load:
                mock_safe_load.return_value = test_schema
                test_lp = LayerProperties(cmf, ".lyr", verify_on_creation=False)

                MapRecipe(fixtures.recipe_with_positive_iso3_code, test_lp)
                self.assertTrue(True, 'validated jsonschema')

        # case where data schema file itself malformed somehow
        with mock.patch('mapactionpy_controller.data_schemas.yaml.safe_load') as mock_safe_load:
            mock_safe_load.return_value = failing_schema

            self.assertRaises(
                jsonschema.exceptions.SchemaError,
                MapRecipe,
                fixtures.recipe_with_positive_iso3_code,
                test_lp
            )

    def test_verify_layer_file_path(self):
        """
        Test the case that a recipe is re-stored with the layer_file_paths already populated, and
        whether or not that path is valid.
        """
        fail_msg = 'The expected layer file'
        with self.assertRaises(ValueError) as arcm:
            MapRecipe(fixtures.recipe_with_invalid_layer_file_path, self.lyr_props)

        if six.PY2:
            self.assertRegexpMatches(str(arcm.exception), fail_msg)
        else:
            self.assertRegex(str(arcm.exception), fail_msg)

    def test_check_lyr_is_in_recipe(self):
        """
        Check that an error is raised if any `step.funcs` attempt to act on a layer object that
        isn't part of the relevant recipe object.
        """
        # Two recipes without a common layer
        recipe_A = MapRecipe(fixtures.recipe_with_layer_name_only, self.lyr_props)
        recipe_B = MapRecipe(fixtures.recipe_with_positive_iso3_code, self.lyr_props)

        test_lyr = recipe_A.all_layers().pop()

        # This should pass without error
        test_lyr._check_lyr_is_in_recipe(recipe_A)

        # This should fail because test_lyr is not from recipe_B
        fail_msg = 'which is not part of the recipe'
        with self.assertRaises(ValueError) as arcm:
            test_lyr._check_lyr_is_in_recipe(recipe_B)

        if six.PY2:
            self.assertRegexpMatches(str(arcm.exception), fail_msg)
        else:
            self.assertRegex(str(arcm.exception), fail_msg)

    def test_calc_layer_file_checksum(self):
        test_recipe = MapRecipe(fixtures.recipe_with_layer_name_only, self.lyr_props)

        test_lyr = test_recipe.all_layers().pop()
        # This is an empty .lyr file
        test_lyr.layer_file_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'test_layer_rendering', 'four_files_exact_match',
            'locationmap_stle_stl_pt_s0_locationmaps.lyr')

        test_lyr._calc_layer_file_checksum()
        hash_of_empty_str = 'd41d8cd98f00b204e9800998ecf8427e'
        self.assertEqual(test_lyr.layer_file_checksum, hash_of_empty_str)

    def test_calc_data_source_checksum(self):
        test_recipe = MapRecipe(fixtures.recipe_with_layer_name_only, self.lyr_props)

        test_lyr = test_recipe.all_layers().pop()
        # Use a simple test shapefile
        test_lyr.data_source_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'test_shp_files',
            'lbn_admn_ad0_py_s1_pp_cdr.shp')

        actual_has_of_shp_file = test_lyr._calc_data_source_checksum()
        expected_hash_of_shp_file = '1acb212b47c8ccb3006ae9b4c5f1cfc0'
        self.assertEqual(actual_has_of_shp_file, expected_hash_of_shp_file)

    def test_get_schema_checker(self):
        test_recipe = MapRecipe(fixtures.recipe_with_layer_name_only, self.lyr_props)

        test_lyr = test_recipe.all_layers().pop()
        # Use a simple test shapefile
        test_lyr.data_source_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'test_shp_files',
            'lbn_admn_ad0_py_s1_pp_cdr.shp')

        passing_schema = yaml.safe_load(r"""
required:
    - admin0Name
    - admin0Na_1
    - admin0Pcod
    - admin0RefN
    - admin0AltN
properties:
    geometry_type:
        items:
            enum:
                - MultiPolygon
                - Polygon
        additionalItems: false
    additionalItems: true
""")

        incorrect_geom_schema = yaml.safe_load(r"""
required:
    - admin0Name
    - admin0Na_1
    - admin0Pcod
    - admin0RefN
    - admin0AltN
properties:
    geometry_type:
        items:
            enum:
                - Point
        additionalItems: false
    additionalItems: true
""")

        incorrect_columns_schema = yaml.safe_load(r"""
required:
    - county_name,
    - parish_name
properties:
    geometry_type:
        items:
            enum:
                - MultiPolygon
                - Polygon
        additionalItems: false
    additionalItems: true
""")

        incorrect_columns_and_geom_schema = yaml.safe_load(r"""
required:
    - county_name,
    - parish_name
properties:
    geometry_type:
        items:
            enum:
                - Point
        additionalItems: false
    additionalItems: true
""")

        # Case where data matches schema
        for test_schema in [passing_schema]:
            test_lyr.data_schema = test_schema
            result = test_lyr.check_data_against_schema(state=test_recipe)
            self.assertTrue(result)

        # Cases where data doesn't match schema
        for test_schema in [incorrect_geom_schema, incorrect_columns_schema, incorrect_columns_and_geom_schema]:
            test_lyr.data_schema = test_schema
            with self.assertRaises(ValueError) as arcm:
                test_lyr.check_data_against_schema(state=test_recipe)

            ve = arcm.exception
            self.assertIsInstance(ve.args[0], recipe_layer.FixSchemaErrorTask)

        # test a logger is call for case where a raster layer is tested. This unittest
        # is only available for py3.4+, though the production code is required for py2.7
        if six.PY3:
            test_lyr.data_source_path = os.path.join(
                self.parent_dir, 'tests', 'testfiles', 'test_shp_files',
                'lbn_evel_dem_ras_s1_pp_cdr.tif')
            test_lyr.data_schema = passing_schema

            with self.assertLogs(level='INFO') as cm:
                result = test_lyr.check_data_against_schema(state=test_recipe)

            print('cm.output = {}'.format(cm.output))
            self.assertRegexpMatches(
                cm.output.pop(),
                'Unable to check schema on for data sources which aren.t shapefiles'
            )

    def test_calc_extent(self):
        test_recipe = MapRecipe(fixtures.recipe_with_layer_name_only, self.lyr_props)

        test_lyr = test_recipe.all_layers().pop()
        # Use a simple test shapefile
        test_lyr.data_source_path = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'test_shp_files',
            'lbn_admn_ad0_py_s1_pp_cdr.shp')

        # Get the extents shapefile and update the test_lyr object
        test_lyr.calc_extent(state=test_recipe)

        expected_extent = [35.10348736558511, 33.054996785738204, 36.62291533501688, 34.69206915371]
        # expected_crs = (
        #     'GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",'
        #     'SPHEROID["WGS_1984",6378137.0,298.257223563]],'
        #     'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]]')
        # expected_crs = {'init': 'epsg:4326'}
        expected_crs = 'epsg:4326'

        for actual, expected in zip(test_lyr.extent, expected_extent):
            self.assertEqual(actual, expected)

        self.assertEqual(test_lyr.crs, expected_crs)



import os
from unittest import TestCase

from mapactionpy_controller.layer_properties import LayerProperties
from mapactionpy_controller.crash_move_folder import CrashMoveFolder


class TestLayerProperties(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.path_to_valid_cmf_des = os.path.join(self.parent_dir, 'example', 'cmf_description_flat_test.json')
        self.path_to_invalid_cmf_des = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_cmf_description_one_file_and_one_dir_not_valid.json')

    def test_different_cmf_args(self):
        # 1) test with valid cmf object
        test_cmf = CrashMoveFolder(self.path_to_valid_cmf_des)
        test_lp = LayerProperties(test_cmf, "test", verify_on_creation=False)
        self.assertIsInstance(test_lp, LayerProperties)

        # 2) test with valid cmf file
        test_lp = LayerProperties(self.path_to_valid_cmf_des, "test", verify_on_creation=False)
        self.assertIsInstance(test_lp, LayerProperties)

        # 3) test with invalid cmf object (eg and CrashMoveFolder object
        #    where verify_paths() returns False)
        test_cmf = CrashMoveFolder(self.path_to_invalid_cmf_des, verify_on_creation=False)
        self.assertRaises(ValueError, LayerProperties, test_cmf, "test")

        # 4) test with invalid cmd file
        self.assertRaises(ValueError, LayerProperties, self.path_to_invalid_cmf_des, "test")

    def test_zero_length_file_extention(self):
        test_cmf = CrashMoveFolder(self.path_to_valid_cmf_des)

        test_lp1 = LayerProperties(test_cmf, '', verify_on_creation=False)
        self.assertIsInstance(test_lp1, LayerProperties)

        layer_rendering_test_root = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'test_layer_rendering')
        test_cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_layer_properties_four_layers.json')
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_exact_match')
        test_lp2 = LayerProperties(test_cmf, '', verify_on_creation=True)
        self.assertIsInstance(test_lp2, LayerProperties)

    def test_verify_with_rendering_files(self):
        # self.fail()

        # load a valid CMF
        test_cmf = CrashMoveFolder(self.path_to_valid_cmf_des)
        # Overwright the Layer Properties file path with
        # mapactionpy_controller\tests\testfiles\fixture_layer_properties_four_layers.json
        test_cmf.layer_properties = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'fixture_layer_properties_four_layers.json')

        layer_rendering_test_root = os.path.join(
            self.parent_dir, 'tests', 'testfiles', 'test_layer_rendering')
        test_cmf.layer_rendering

        # 1) Exact match of .lyr files and layer properties
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_exact_match')
        lyr_lp = LayerProperties(test_cmf, '.lyr', verify_on_creation=True)
        self.assertFalse(lyr_lp.is_difference_with_layer_rendering_dir())

        qml_lp = LayerProperties(test_cmf, '.qml', verify_on_creation=True)
        self.assertFalse(qml_lp.is_difference_with_layer_rendering_dir())

        # 2) .lyr files which don't have layer properties entries
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'five_files')
        self.assertRaises(ValueError, LayerProperties, test_cmf, '.lyr')
        self.assertRaises(ValueError, LayerProperties, test_cmf, '.qml')

        # 3) layer properties entries which don't have cooresponding .lyr files.
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'three_files')
        self.assertRaises(ValueError, LayerProperties, test_cmf, '.lyr')
        self.assertRaises(ValueError, LayerProperties, test_cmf, '.qml')

        # 4) Both 2 & 3 combined
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_mis_match')
        self.assertRaises(ValueError, LayerProperties, test_cmf, '.lyr')
        self.assertRaises(ValueError, LayerProperties, test_cmf, '.qml')

        # 6) Overrided validation checks in constructor
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_mis_match')
        long_lived_lp = LayerProperties(test_cmf, '.lyr', verify_on_creation=False)
        self.assertTrue(long_lived_lp.is_difference_with_layer_rendering_dir())
        long_lived_lp.cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_exact_match')
        self.assertFalse(long_lived_lp.is_difference_with_layer_rendering_dir())

        # 7) Same results whether or not the '.' character is included in the file extension
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_exact_match')
        # passing example with dot
        lyr_lp = LayerProperties(test_cmf, '.lyr', verify_on_creation=True)
        self.assertFalse(lyr_lp.is_difference_with_layer_rendering_dir())
        # passing example without dot
        lyr_lp = LayerProperties(test_cmf, 'lyr', verify_on_creation=True)
        self.assertFalse(lyr_lp.is_difference_with_layer_rendering_dir())
        # failing example with dot
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_mis_match')
        self.assertRaises(ValueError, LayerProperties, test_cmf, '.lyr')
        # failing example without dot
        test_cmf.layer_rendering = os.path.join(layer_rendering_test_root, 'four_files_mis_match')
        self.assertRaises(ValueError, LayerProperties, test_cmf, 'lyr')



import os
from unittest import TestCase, skip

import mapactionpy_controller.task_renderer as task_renderer
from mapactionpy_controller.name_convention import NamingConvention
from mapactionpy_controller.task_renderer import FixDataNameTask, TaskReferralBase
from mapactionpy_controller.crash_move_folder import CrashMoveFolder
from mapactionpy_controller.event import Event


class TestTaskRendering(TestCase):

    def setUp(self):
        self.parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.dir_to_valid_cmf_des = os.path.join(self.parent_dir, 'example')
        self.path_to_valid_cmf_des = os.path.join(self.dir_to_valid_cmf_des, 'cmf_description_flat_test.json')
        self.path_to_valid_event_des = os.path.join(self.dir_to_valid_cmf_des, 'event_description.json')
        self.hum_event = Event(self.path_to_valid_event_des)

    def test_get_task_unique_summary(self):
        dnc_json_path = os.path.join(self.parent_dir, 'example', 'data_naming_convention.json')
        nc = NamingConvention(dnc_json_path)
        cmf = CrashMoveFolder(self.path_to_valid_cmf_des)

        # Three cases
        # 1. Parsable and Valid
        # 2. Parsable and Not Valid
        # 3. Not Parsable
        #
        # key= is the name to test
        # value= is a list of expected results from rendering the mustache elements in `mustache_tmpls`
        test_cases = {
            'lka_admn_ad2_py_s5_unocha_pp_freetext': [
                'lka_admn_ad2_py_s5_unocha_pp_freetext',
                'parsable but not valid',
                (
                    '* "ad2" is a recognised value for the clause "datatheme" found in "03_theme.csv"\n'
                    '* "admn" is a recognised value for the clause "datacat" found in "02_category.csv"\n'
                    '* "freetext" is valid for freetext (as is almost anything)\n'
                    '* "lka" is a recognised value for the clause "geoext" found in "01_geoextent.csv"\n'
                    '* "pp" is a recognised value for the clause "perm" found in "07_permission.csv"\n'
                    '* "py" is a recognised value for the clause "geom" found in "04_geometry.csv"\n'
                    '* "s5" is a recognised value for the clause "scale" found in "05_scale.csv"\n'
                    '* "unocha" is a recognised value for the clause "source" found in "06_source.csv"\n'
                ),
                ''
            ],
            'lka_admn_ad77_py_s5_unocha_pp_freetext': [
                'lka_admn_ad77_py_s5_unocha_pp_freetext',
                'parsable but not valid',
                (
                    '* "admn" is a recognised value for the clause "datacat" found in "02_category.csv"\n'
                    '* "freetext" is valid for freetext (as is almost anything)\n'
                    '* "lka" is a recognised value for the clause "geoext" found in "01_geoextent.csv"\n'
                    '* "pp" is a recognised value for the clause "perm" found in "07_permission.csv"\n'
                    '* "py" is a recognised value for the clause "geom" found in "04_geometry.csv"\n'
                    '* "s5" is a recognised value for the clause "scale" found in "05_scale.csv"\n'
                    '* "unocha" is a recognised value for the clause "source" found in "06_source.csv"\n'
                ),
                '* "ad77" is not a recognised value for the clause "datatheme" found in "03_theme.csv"\n'
            ],
            'unparsable-name': [
                'unparsable-name',
                'not parsable',
                '',
                ''
            ]
        }

        mustache_tmpls = [
            '<%name_result.name_to_validate%>',
            ('<%#name_result.is_parsable%>parsable but not valid<%/name_result.is_parsable%>'
             '<%^name_result.is_parsable%>not parsable<%/name_result.is_parsable%>'),
            '<%#name_result.valid_clause_list%>* <%&valid_clause%>\n<%/name_result.valid_clause_list%>',
            '<%#name_result.invalid_clause_list %>* <%&invalid_clause%>\n<%/name_result.invalid_clause_list %>'
        ]

        # for name_to_test, expected_result_list in test_cases.iteritems():
        while test_cases:
            name_to_test, expected_result_list = test_cases.popitem()
            # print(name_to_test)
            nr = nc.validate(name_to_test)
            fdnt = FixDataNameTask(nr, cmf)

            for test_template, expected_result in zip(mustache_tmpls, expected_result_list):
                # create the Task object
                # override the unique identifier for test purposes
                fdnt._primary_key_template = test_template

                actual_result = fdnt.get_task_unique_summary()
                self.assertEqual(actual_result, expected_result)

    def test_get_task_description(self):
        trb = TaskReferralBase(self.hum_event)
        self.assertIn('Major Configuration Error', trb.get_task_description())

    def test_cmf_description_adapter(self):
        test_cmf = CrashMoveFolder(self.path_to_valid_cmf_des)
        test_cd = task_renderer.cmf_description_adapter(test_cmf)
        self.assertEqual(self.dir_to_valid_cmf_des, test_cd['cmf']['path'])

    @skip('Not ready yet')
    def test_render_with_schema_error(self):
        self.fail()

    @skip('Not ready yet')
    def test_render_with_missing_file(self):
        self.fail()

    @skip('Not ready yet')
    def test_render_with_misplaced_file(self):
        self.fail()



import json
import os
import pycountry
from mapactionpy_controller import _get_validator_for_config_schema

validate_against_event_schema = _get_validator_for_config_schema('event-v0.2.schema')


class Event:
    def __init__(self, event_file):

        self.path = os.path.dirname(event_file)
        with open(event_file, 'r') as f:
            event_def = json.loads(f.read())
            validate_against_event_schema(event_def)

            # Doubtless there is a more elegant way to do this.
            # 1x file path
            self.cmf_descriptor_path = os.path.join(self.path, event_def['cmf_descriptor_path'])
            # 3x integers
            self.default_jpeg_res_dpi = int(event_def['default_jpeg_res_dpi'])
            self.default_pdf_res_dpi = int(event_def['default_pdf_res_dpi'])
            self.default_emf_res_dpi = int(event_def['default_emf_res_dpi'])
            # 12x others
            self.operation_name = event_def['operation_name']
            self.glide_number = event_def['glide_number']
            self.affected_country_iso3 = event_def['affected_country_iso3'].lower()
            self.time_zone = event_def['time_zone']
            self.language_iso2 = event_def['language_iso2']
            self.operation_id = (event_def['operation_id']).lower()
            self.default_source_organisation = event_def['default_source_organisation']
            self.default_source_organisation_url = event_def['default_source_organisation_url']
            self.default_publishing_base_url = event_def['default_publishing_base_url']
            self.deployment_primary_email = event_def['deployment_primary_email']
            self.default_disclaimer_text = event_def['default_disclaimer_text']
            self.default_donor_credits = event_def['default_donor_credits']
            # self.donors = event_def['donors']
            self.country_name = _parse_country_name(event_def)


def _parse_country_name(event_def):
    """
    Checks the optional `country_name` value of the event description file and set the cooresponsing
    value in the Event object. Countries may be real or fictional, but both the `affected_country_iso3`
    and `country_name` values must match.

    * If the `affected_country_iso3` value refers to a real country, then the `country_name` key:value
        must fulfill one of these critica:
        * Not exist
        * When searched using `pycountry.countries.search_fuzzy`, one of the countries returned must
          have the correct ISO3 code to match `affected_country_iso3`.

    * If the `affected_country_iso3` value refers to a real country and the `country_name` key:value
        is specified, then its value must not match the "name" or "official_name" of any real country
        (as given by `pycountry`).

    * If the `affected_country_iso3` value refers to a real country and the `country_name` key:value
        is not specified, then the Event object will be created with the member
        `country_name=='Unknown country'`.

    @param event_def: A dict repesenting the json event description
    @returns: The validated Country Name. If a value is given in the JSON, and the tests pass then
              value as specified in JSON is returned, even if it is not identical to the `name` or
              `official_name` given by pycountry.
    @raises ValueError: In either of these two circumstainces:
        * If event_def['affected_country_iso3'] is a real country ISO3 code
            AND event_def['country_name'] is a fictional name.
        * If event_def['affected_country_iso3'] is a fictional country ISO3 code
            AND event_def['country_name'] is (or is close to) a real country name.
    """

    # Look up Country Name in pycountry
    affected_country_iso3 = event_def['affected_country_iso3'].upper()
    lookup_ctry = pycountry.countries.get(alpha_3=affected_country_iso3)
    raw_name = event_def.get('country_name', None)
    validation_results = None
    try:
        validation_results = pycountry.countries.search_fuzzy(raw_name)
    except (LookupError, AttributeError):
        pass

    # Real ISO
    if lookup_ctry:
        return _parse_real_country_name(raw_name, validation_results, lookup_ctry, affected_country_iso3)

    # Fictional ISO
    return _parse_fictional_country_name(raw_name, validation_results, affected_country_iso3)


def _parse_fictional_country_name(raw_name, validation_results, affected_country_iso3):
    if validation_results:
        # Fictional ISO, Real Country name - BAD
        real_ctrys = ', '.join([country.name for country in validation_results])
        raise ValueError('It is not valid to supply a real country name with a fictional ISO3 code.'
                         ' Fictional value for affected_country_iso3="{}". Supplied country_name="{}".'
                         ' Simular real country name(s)="{}".'.format(
                             affected_country_iso3, raw_name, real_ctrys))

    # Fictional ISO, No Country name - BAD
    if raw_name is None:
        raise ValueError('A `country_name` value must be specified for fictional countries.'
                         ' Fictional value for affected_country_iso3="{}"'.format(affected_country_iso3))

    # Fictional ISO, Fictional Country name - OK
    return raw_name


def _parse_real_country_name(raw_name, validation_results, lookup_ctry, affected_country_iso3):
    if validation_results and len(validation_results) > 0:
        if lookup_ctry.alpha_3 in [country.alpha_3 for country in validation_results]:
            # Real ISO, Real Country name - OK
            return raw_name

    # Real ISO, No Country name - OK
    if raw_name is None:
        return lookup_ctry.name

    # Real ISO, Fictional Country name
    raise ValueError('It is not valid to supply a real ISO3 code with a fictional country name.'
                     ' Supplied affected_country_iso3="{}". Supplied fictional country_name="{}".'
                     ' Real country name determined from supplied ISO3 code="{}"'.format(
                         affected_country_iso3, raw_name, lookup_ctry.name))



# Required to generate Export XML
# Mapdata in its own class so that it's held in the inner mapdata block:
#
# <?xml version="1.0" ?>
# <mapdoc>
#    <mapdata>


class MapData:
    def __init__(self, row):
        # Constructor.  Creates an instance of MapData for the Export XML

        self.versionNumber = row["versionNumber"]
        self.mapNumber = row["mapNumber"]
        self.operationID = row["operationID"]
        self.sourceorg = row["sourceorg"]
        self.glideno = row["glideno"]

        self.jpgfilename = row["jpgfilename"]
        self.pdffilename = row["pdffilename"]
        self.emffilename = row.get("emffilename", "")
        self.jpgfilesize = row["jpgfilesize"]
        self.pdffilesize = row["pdffilesize"]
        self.emffilesize = row.get("emffilesize", '')

        self.title = row["title"]
        self.ref = row["ref"]
        self.language = row["language"]
        self.countries = row["countries"]
        self.createdate = row["createdate"]
        self.createtime = row["createtime"]
        self.status = row["status"]
        self.xmin = row["xmin"]
        self.ymin = row["ymin"]
        self.xmax = row["xmax"]
        self.ymax = row["ymax"]
        self.proj = row["proj"]
        self.datum = row["datum"]
        self.qclevel = row["qclevel"]
        self.qcname = row["qcname"]
        self.access = row["access"]
        self.summary = row["summary"]
        self.imagerydate = row["imagerydate"]
        self.datasource = row["datasource"]
        self.location = row["location"]
        self.themes = row["themes"]
        self.scale = row["scale"]
        self.papersize = row["papersize"]
        self.jpgresolutiondpi = row["jpgresolutiondpi"]
        self.pdfresolutiondpi = row["pdfresolutiondpi"]
        self.kmlresolutiondpi = row["kmlresolutiondpi"]
        self.mxdfilename = row["mxdfilename"]
        self.paperxmax = row["paperxmax"]
        self.paperxmin = row["paperxmin"]
        self.paperymax = row["paperymax"]
        self.paperymin = row["paperymin"]
        self.accessnotes = row["accessnotes"]
        self.product_type = row["product-type"]  # Name contains hyphen "-"
        self.language_iso2 = row["language-iso2"]  # Name contains hyphen "-"
